import json
import os
import socket
from platform import uname
from shutil import which
from typing import List, Optional, Tuple

import click

from vessl.openapi_client.models import ResponseKernelClusterInfo
from vessl.cli._base import VesslGroup, vessl_argument, vessl_option
from vessl.cli._util import (
    print_data,
    print_table,
)
from vessl.util.fmt import format_string
from vessl.util.prompt import prompt_confirm, prompt_choices, generic_prompter
from vessl.util.echo import print_warning, print_debug, print_info, print_success, print_success_result, print_error, \
    print_error_result
from vessl.util.endpoint import Endpoint
from vessl.cli.organization import organization_name_option
from vessl.kernel_cluster import (
    CreateClusterParam,
    _acquire_kubeconfig,
    _check_existing_vessl_installation,
    _get_running_k0s_container,
    _load_cluster_kubeconfig,
    _load_kubeconfig_to_config,
    _start_k0s_container,
    _wait_for_k0s_ready,
    add_and_update_helm_repo,
    create_cluster,
    delete_cluster,
    list_cluster_nodes,
    list_clusters,
    read_cluster,
    rename_cluster,
)
from vessl.organization import _get_organization_name
from vessl.util.config import VesslConfigLoader
from vessl.util.constant import VESSL_HELM_REPO
from vessl.util.exception import (
    ACCESS_TOKEN_NOT_FOUND_ERROR_CODE,
    InvalidKernelClusterError,
    VesslApiException,
)


def cluster_name_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    clusters = list_clusters()
    return prompt_choices("Cluster", [x.name for x in clusters])


def custom_cluster_name_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    clusters = list_clusters()
    clusters = [x for x in clusters if not x.is_savvihub_managed]
    return prompt_choices("Cluster", [x.name for x in clusters])


@click.command(name="cluster", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
@vessl_option(
    "--name",
    type=click.STRING,
    required=True,
    help="Name of the cluster. (e.g. 'seoul-cluster')",
    prompt="Enter cluster name",
)
@vessl_option(
    "--kubernetes-namespace",
    type=click.STRING,
    default="vessl",
    help="Kubernetes namespace to install VESSL agent (default: vessl)",
    prompt="Enter Kubernetes namespace to install VESSL agent",
    prompt_required=True,
)
@vessl_option(
    "--provider",
    type=click.Choice(["aws", "gcp", "on-premise"], case_sensitive=False),
    default="on-premise",
    help="Cluster provider; aws|gcp|on-premise (default: on-premise)",
    prompt="Select your cluster provider",
    prompt_required=True,
)
@vessl_option(
    "--extra-helm-values",
    type=click.STRING,
    multiple=True,
    default=[],
    help="Extra Helm values to pass when installing VESSL agent.",
)
@organization_name_option
def create(
    name: str,
    kubernetes_namespace: str,
    provider: str,
    extra_helm_values: List[str],
    organization_name: str = None,
):
    ########################################
    # validate inputs
    ########################################
    if name is None or len(name) < 4:
        print_error_result("Cluster name should be at least 4 characters long")
        return

    ########################################
    # validate pre-creation conditions
    ########################################
    if _check_helm_availability() == False:
        return
    print_info("Cluster installation dependencies are met.")

    config_path, context = _try_load_kubeconfig()
    if config_path is None:
        return
    print_info("Kubernetes configuration is loaded")
    print_data({"Kubeconfig path": config_path, "Current kubernetes context": context})

    if _vessl_already_exist(kubernetes_namespace):
        if not prompt_confirm(
            f"VESSL agent is already installed on current Kubernetes cluster. Do you want to remove and re-install?",
            default=False,
        ):
            return
        else:
            print_info("Cluster already has VESSL agent installed but force continuing...")
            print_warning("Removing existing VESSL agent...")
            remove_success = _try_remove_vessl_agent(kubernetes_namespace)
            if not remove_success:
                return

    else:
        print_info("VESSL cluster not found in your environment. Installing VESSL cluster...")

    ########################################
    # prepare helm repo
    ########################################
    if not prompt_confirm(
        f"Adding VESSL helm repo: {VESSL_HELM_REPO}. Continue?",
        default=True,
    ):
        print_error_result("Cluster creation aborted")
        print_info("You can add VESSL helm repo manually by running:")
        print_info(f"helm repo add vessl {VESSL_HELM_REPO}")
        return
    add_and_update_helm_repo()

    input_data = {
        "VESSL cluster name": name,
        "Provider": provider,
        "Kubeconfig path": config_path,
        "Current kubernetes context": context,
        "Kubernetes namespace": kubernetes_namespace,
    }

    ########################################
    # build create cluster param
    ########################################
    create_cluster_param = _build_create_cluster_param(
        name,
        kubernetes_namespace,
        provider,
        extra_helm_values,
        config_path,
        organization_name,
        input_data,
    )

    ########################################
    # user confirmation
    ########################################
    print_info("Cluster creation options:")
    print_data(input_data)
    if not prompt_confirm(
        f"Are you sure you want to install VESSL agent on current Kubernetes cluster with options displayed above?",
        default=True,
    ):
        print_error_result("Cluster creation aborted")
        return

    print_info("Creating cluster...")
    vessl_installation_info = create_cluster(create_cluster_param)
    VesslConfigLoader().kubernetes_namespace = kubernetes_namespace
    print_success_result(f"cluster {name} created!")
    print_cluster_data(vessl_installation_info)


def _build_create_cluster_param(
    name,
    kubernetes_namespace,
    provider,
    extra_helm_values,
    config_path,
    organization_name,
    input_data,
) -> CreateClusterParam:
    """Builds CreateClusterParam object from user inputs.\n
    Conditionally adds storage class, local path provisioner, and pod resource path.\n
    Also, sets input data for user confirmation.

    """
    create_cluster_param = CreateClusterParam()
    create_cluster_param.cluster_name = name
    create_cluster_param.kubernetes_namespace = kubernetes_namespace
    create_cluster_param.provider = provider
    create_cluster_param.kubeconfig_path = config_path
    if organization_name:
        create_cluster_param.organization_name = organization_name

    if provider == "gcp":
        print_debug(
            f"(provider=gcp) Setting default storage class for GCP as [standard-rwo]",
        )
        input_data["default storage class (gcp)"] = "standard-rwo"
    if provider == "aws":
        print_debug(
            f"(provider=aws) Setting default storage class for AWS as [vessl-ebs]",
        )
        input_data["default storage class (aws)"] = "vessl-ebs"

    # FIXME(mika): 추후 local path provisioner가 모든 클러스터에 마이그레이션 되면 아래의 값들을 전부 enable_longhorn으로 바꾼다.
    should_enable_local_path_provisioner = provider == "on-premise"
    print_debug(
        f"{'Enable' if should_enable_local_path_provisioner else 'Disable'} local-path-provisioner for {provider} cluster.\n"
        + "  This will allow VESSL agent to use host local storage for your workspace.\n"
    )
    input_data[f"local-path-provisioner"] = should_enable_local_path_provisioner
    if not should_enable_local_path_provisioner:
        # Default enabled for on-prem, disabled for cloud
        create_cluster_param.enable_local_path_provisioner = False
        # FIXME(mika): mitigation for non on-prem cluster installs.
        create_cluster_param.enable_longhorn = False

    if extra_helm_values and len(extra_helm_values) > 0:
        print_debug(
            f"Extra Helm values will be passed to VESSL agent installation: {extra_helm_values}",
        )
        input_data["Extra Helm values"] = ", ".join(extra_helm_values)
        create_cluster_param.extra_helm_values = extra_helm_values

    # dcgm patch
    from kubernetes import client as k8s_client

    v1 = k8s_client.CoreV1Api()
    # @XXX(seokju) rough assumption: k0s user will install k0s nodes
    # if this becomes a problem, patch chart and cli.
    node = v1.list_node().items[0]
    if "k0s" in node.status.node_info.kubelet_version:
        print_debug(
            f"(Required) Detected k0s node. Patching dcgm exporter podResourcePath as /var/lib/k0s/kubelete/pod-resources.",
        )
        input_data["(k0s) Patch dcgm exporter podResourcePath"] = True
        create_cluster_param.pod_resource_path = "/var/lib/k0s/kubelet/pod-resources"
    return create_cluster_param


def _vessl_already_exist(namespace: str):
    print_info("Checking existing VESSL installation...")
    try:
        vessl_installation_info = _check_existing_vessl_installation(namespace=namespace)
        if vessl_installation_info is not None:
            print_success("Existing VESSL installation found.")
            print_cluster_data(vessl_installation_info)
            return True
        return False
    except InvalidKernelClusterError as e:
        if e.code == ACCESS_TOKEN_NOT_FOUND_ERROR_CODE:
            print_error_result("Existing VESSL installation found, but access token is missing.")
        else:
            print_error_result(f"Existing VESSL installation found, but it is invalid. {repr(e)}")
        return True
    except VesslApiException as e:
        if e.status == 404:
            print_error(
                f"Existing VESSL installation, but not registered in VESSL Service: [{repr(e)}]"
            )
        else:
            print_error_result(
                f"Existing VESSL installation, but failed to check registration on VESSL Service: {repr(e)}"
            )
        return True


def _try_remove_vessl_agent(namespace: str) -> bool:
    config_path, context = _try_load_kubeconfig()
    if config_path is None:
        print_error_result("Failed to load kubeconfig")
        return False
    stream = os.popen(f"helm uninstall vessl --namespace {namespace} --kubeconfig {config_path}")
    output = stream.read()
    print_info(output)
    if stream.close() != None:
        print_error_result("Failed to uninstall vessl-agent")
        return False
    return True


def _try_load_kubeconfig():
    print_info("Loading cluster kubeconfig...")
    loaded, (config_path, context) = _load_cluster_kubeconfig()
    if not loaded:
        print_error_result("Kubernetes cluster not found;")
        print_error("Please configure valid kubeconfig file location by KUBECONFIG envvar.")
        print_info(
            "If you are a test user and want to install k0s-in-docker on your local machine, please run 'vessl cluster init-local' first."
        )
        return None, None

    return config_path, context


def _check_helm_availability():
    print_debug("Checking if helm exists...")
    if which("helm") is None:
        print_error_result(
            "Command 'helm' not found; Please install Helm to install VESSL cluster agent: https://helm.sh/"
        )
        return False
    print_debug("helm command available!")
    return True


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=cluster_name_prompter)
@organization_name_option
def read(name: str):
    cluster = read_cluster(cluster_name=name)
    print_cluster_data(cluster)


@cli.vessl_command()
@organization_name_option
def list():
    clusters = list_clusters()
    print_table(
        clusters,
        ["ID", "Name", "Type", "Status", "K8s Master Endpoint", "K8s Namespace"],
        lambda x: [
            x.id,
            x.name,
            "Managed" if x.is_savvihub_managed else "Custom",
            x.status.replace("-", " "),
            format_string(x.name),
            format_string(x.kubernetes_namespace),
        ],
    )


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=custom_cluster_name_prompter)
@organization_name_option
def delete(name: str):
    if not prompt_confirm(f"Are you sure you want to delete '{name}'?"):
        print_error_result("Cluster deletion aborted")
        return
    cluster = read_cluster(cluster_name=name)
    data = delete_cluster(cluster_id=cluster.id)
    print_success(f"Deleted '{name}'.")

    config = VesslConfigLoader()

    if prompt_confirm(f"Do you want to destroy the cluster installed with VESSL CLI?"):
        namespace = config.kubernetes_namespace
        if namespace is None:
            print_error_result(
                "Failed to find kubernetes namespace. Please check your kubeconfig file."
            )
        else:
            if not _try_remove_vessl_agent(namespace=namespace):
                print_error(
                    "Failed to remove vessl-agent. Please remove it manually by 'helm uninstall vessl --namespace <namespace>'"
                )
            else:
                print_success("vessl-agent removed successfully.")

    if config.use_k0s_in_docker:
        if _check_k0s_exists():
            print_info(
                "It seems that you are using k0s-in-docker installed by `vessl cluster init-local`."
            )
            if prompt_confirm(
                "Do you want to remove k0s-in-docker? (note that this is NOT RECOVERABLE)"
            ):
                is_error = _try_remove_existing_k0s()
                if is_error:
                    print_error_result("Failed to remove k0s-in-docker.")
                    print_info(
                        f"You can try execute the following in the shell (note that this is NOT RECOVERABLE)."
                    )
                    print_info(f"$ docker stop vessl-k0s && docker rm vessl-k0s")
                    print_info(
                        f'$ docker run -it --rm --privileged --pid=host alpine nsenter -t 1 -m -- sh -c "rm -rfv /var/lib/k0s"'
                    )
        else:
            config.use_k0s_in_docker = False


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=custom_cluster_name_prompter)
@vessl_argument("new_name", type=click.STRING, required=True, prompter=generic_prompter("New name"))
@organization_name_option
def rename(name: str, new_name: str):
    cluster = read_cluster(cluster_name=name)
    cluster = rename_cluster(cluster_id=cluster.id, new_cluster_name=new_name)
    print_success(f"Renamed '{name}' to '{cluster.name}'.")


@cli.vessl_command()
@vessl_argument("name", type=click.STRING, required=True, prompter=custom_cluster_name_prompter)
@organization_name_option
def list_nodes(name: str):
    cluster = read_cluster(cluster_name=name)
    cluster_nodes = list_cluster_nodes(cluster_id=cluster.id)
    print_table(
        cluster_nodes,
        ["ID", "Name", "CPU Limits", "GPU Limits", "Memory Limits"],
        lambda x: [x.id, x.name, x.cpu_limits, x.gpu_limits, x.memory_limits],
    )


@cli.vessl_command()
def init_local():
    print_info("Checking if docker exists...")
    import docker

    try:
        docker.from_env()
    except docker.errors.DockerException as e:
        print_error(
            f"[{repr(e)}] Docker not found; Please install Docker to run VESSL on your machine.: https://docs.docker.com/engine/install/"
        )
        return
    print_info("docker command available!")

    print_info("Checking if k0s-in-docker already exists...")
    if _check_k0s_exists():
        if prompt_confirm("k0s-in-docker already exists. Do you want to remove and reinstall it?"):
            is_error = _try_remove_existing_k0s()
            if is_error:
                print_error_result("Failed to remove k0s-in-docker.")
                return
            print_success("k0s-in-docker removed successfully.")
        else:
            print_info("You need to remove existing local cluster first.")
            print_info("Aborted.")
            return

    if not prompt_confirm(
        "Are you sure you want to initialize local cluster with k0s on docker?",
    ):
        print_info("Aborted.")
        return

    print_info(
        "Creating local cluster with k0s... (More Info: https://docs.k0sproject.io/v1.25.12+k0s.0/k0s-in-docker)"
    )
    kubeconfig, context_name, already_exists = _install_k0s_in_docker()
    if kubeconfig is None:
        print_error("Failed to create local cluster!")
        return
    if already_exists:
        print_info("Cluster already exists")
    else:
        print_info("Local cluster created!")

    VesslConfigLoader().use_k0s_in_docker = True
    print_success(
        "Initializing cluster success!, Hit `vessl cluster create` to rocket your Machine Learning!",
    )


def _check_k0s_exists():
    import docker

    client = docker.from_env()
    containers = client.containers.list(filters={"name": "vessl-k0s"}, all=True)
    if len(containers) > 0:
        print_info("k0s-in-docker already exists!")
        print_data(
            {"Id": containers[0].id, "Name": containers[0].name, "Status": containers[0].status}
        )
        return True
    print_info("k0s-in-docker does not exist!")
    return False


def _try_remove_existing_k0s():
    import docker

    client = docker.from_env()
    containers = client.containers.list(filters={"name": "vessl-k0s"}, all=True)
    is_error = False
    print_info("Removing k0s-in-docker...")
    try:
        for container in containers:
            if container.status == "running":
                container.stop()
            container.remove(force=True)
    except docker.errors.APIError as e:
        print_error(f"Failed to uninstall k0s-in-docker!: {repr(e)}")
        is_error = True
        return is_error
    print_info("Removing connected volume...")
    try:
        stream = os.popen(
            'docker run -it --rm --privileged --pid=host alpine nsenter -t 1 -m -- sh -c "rm -rfv /var/lib/k0s"'
        )
        output = stream.read()
        if stream.close() is not None:
            raise Exception(output)
        print_info(output)
    except Exception as e:
        print_error(f"Failed to remove connected volume!: {repr(e)}")
        print_info("You can try run following command manually;")
        print_info(
            '> docker run -it --rm --privileged --pid=host alpine nsenter -t 1 -m -- sh -c "rm -rfv /var/lib/k0s"'
        )
        is_error = True
        return is_error
    print_info("Uninstalled k0s-in-docker!")

    print_info("Removing custom kubeconfig...")
    try:
        kubeconfig_path = VesslConfigLoader().cluster_kubeconfig
        if os.path.exists(kubeconfig_path):
            os.remove(kubeconfig_path)
    except Exception as e:
        print_error(f"Failed to remove custom kubeconfig!: {repr(e)}")
        is_error = True
        return is_error
    print_info("Removed custom kubeconfig!")

    return is_error


def _install_k0s_in_docker() -> Tuple[Optional[str], Optional[str], bool]:
    """Install k0s-in-Docker on local machine.
    More Info: https://docs.k0sproject.io/v1.25.12+k0s.0/k0s-in-docker/
    """
    import docker

    print_info(f"Installing k0s-in-docker...")
    k0s_config = """apiVersion: k0s.k0sproject.io/v1beta1
kind: ClusterConfig
metadata:
  name: k0s
spec:
  api:
    extraArgs:
      service-node-port-range: 32639-32767
"""
    kwargs = {
        "name": "vessl-k0s",
        "environment": {"K0S_CONFIG": k0s_config},
        "hostname": f"vessl-k0s-{socket.gethostname()}",
        "ports": {
            "6443/tcp": 6443,
        },
        "privileged": True,
        "volumes": {
            "/var/lib/k0s": {
                "bind": "/var/lib/k0s",
                "mode": "rw",
            },
        },
    }
    for i in range(32639, 32768):
        kwargs["ports"][f"{i}/tcp"] = i

    # Add extra arguments for arm64 machines
    if uname().machine == "arm64":
        print_info("System is using CPU with arm64 architecture; adding additional kwargs")
        kwargs["environment"]["ETCD_UNSUPPORTED_ARCH"] = "arm64"

    client = docker.from_env()

    # Add extra arguments for docker container running on Docker Desktop
    is_docker_desktop = client.info().get("OperatingSystem", "") == "Docker Desktop"
    if is_docker_desktop:
        print_info("System is running on Docker Desktop")

    if prompt_confirm(
        f"Should mount cgroup from host? (default: {'Y' if is_docker_desktop else 'N'}, Recommended for Docker Desktop)",
        default=is_docker_desktop,
    ):
        kwargs["cgroupns"] = "host"
        kwargs["volumes"]["/sys/fs/cgroup"] = {
            "bind": "/sys/fs/cgroup",
            "mode": "rw",
        }

    already_exists = False

    # Run k0s container
    print_info("k0s container configuration: ")
    print_warning(json.dumps(kwargs, indent=2))
    if not prompt_confirm("Starting k0s in docker with upper configuraion; Continue?"):
        print_info("Aborted.")
        return None, None, already_exists
    container = None
    try:
        container = _start_k0s_container(kwargs)
    except docker.errors.APIError as e:
        if e.status_code == 409 and "vessl-k0s" in e.explanation:
            print_error(
                "k0s-in-docker container already exists.",
            )
            already_exists = True
        else:
            print_error_result(
                f"[{repr(e)}] Failed to start k0s-in-docker container. Please check your docker configuration.",
            )
            return None, None, already_exists
    if already_exists:
        try:
            container = _get_running_k0s_container()
        except:
            print_error_result(
                "Failed to get running k0s-in-docker container. Please remove the existing container and try again.",
            )
            click.Abort()
            return None, None, already_exists

    # Wait for the container to be ready
    print_info(f"Waiting for the k0s-in-docker node to be ready...")
    try:
        _wait_for_k0s_ready(container)
    except Exception as e:
        click.secho(
            f"[{repr(e)}]; Error while waiting. Failed to initialize k0s-in-docker node.",
            bg="red",
            err=True,
        )
        click.Abort()
        return None, None, already_exists
    print_info(f"k0s-in-docker node is ready!")

    # Get kubeconfig, keep at /opt/vessl/cluster_kubeconfig.yaml for later
    print_info(f"Acquiring kubeconfig...")
    kubeconfig_filename = _acquire_kubeconfig(container)
    current_context = _load_kubeconfig_to_config(kubeconfig_filename)
    print_info(
        f"k0s-in-docker is ready! Saved kubeconfig at {kubeconfig_filename} for later use.",
    )
    return kubeconfig_filename, current_context["name"], already_exists


def cluster_id_prompter(ctx: click.Context, param: click.Parameter, value: str) -> int:
    clusters = list_clusters()
    cluster = prompt_choices(
        "Cluster",
        [(f"{x.name}", x) for x in clusters],
    )
    ctx.obj["cluster"] = cluster
    return cluster.id


def cluster_name_prompter(ctx: click.Context, param: click.Parameter, value: str) -> str:
    clusters = list_clusters()
    cluster = prompt_choices(
        "Cluster",
        [(f"{x.name}", x) for x in clusters],
    )
    ctx.obj["cluster"] = cluster
    return cluster.name


def cluster_name_callback(ctx: click.Context, param: click.Parameter, value: str) -> str:
    if "cluster" not in ctx.obj:
        ctx.obj["cluster"] = read_cluster(value)
    return value


def print_cluster_data(cluster: ResponseKernelClusterInfo, **kwargs):
    print_data(
        {
            "Name": cluster.name,
            "Dashboard URL": Endpoint.cluster.format(_get_organization_name(**kwargs), cluster.id),
            "Type": "VESSL-managed" if cluster.is_savvihub_managed else "Custom",
            "Region": format_string(cluster.region),
            "Status": cluster.status.replace("-", " "),
            "K8s Master Endpoint": format_string(cluster.name),
            "K8s Namespace": format_string(cluster.kubernetes_namespace),
            "K8s Service Type": cluster.kubernetes_service_type,
        }
    )


cluster_option = vessl_option(
    "-c",
    "--cluster",
    type=click.STRING,
    required=True,
    prompter=cluster_name_prompter,
    callback=cluster_name_callback,
    help="Must be specified before resource-related options (`--resource`, `--processor`, ...).",
)



# ruff: noqa

import warnings
warnings.warn("""
################################
###       ! WARNING !        ###
###    DEPRECATION NOTICE    ###
################################

You have imported `openapi_client` package provided by VESSL SDK/CLI.

This package will be DEPRECATED in future releases,
and programs relying on it will become broken.

We strongly recommend that you find and remove imports to `openapi_client`,
and instead use VESSL SDK functions for such purpose.
This package is for internal use; please do not use this directly.
""")
del warnings

from vessl.openapi_client import *


# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmCouponRedemptionEdges(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'coupon': 'OrmCoupon',
        'organization': 'OrmOrganization',
        'user': 'OrmUser'
    }

    attribute_map = {
        'coupon': 'coupon',
        'organization': 'organization',
        'user': 'user'
    }

    def __init__(self, coupon=None, organization=None, user=None, local_vars_configuration=None):  # noqa: E501
        """OrmCouponRedemptionEdges - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._coupon = None
        self._organization = None
        self._user = None
        self.discriminator = None

        if coupon is not None:
            self.coupon = coupon
        if organization is not None:
            self.organization = organization
        if user is not None:
            self.user = user

    @property
    def coupon(self):
        """Gets the coupon of this OrmCouponRedemptionEdges.  # noqa: E501


        :return: The coupon of this OrmCouponRedemptionEdges.  # noqa: E501
        :rtype: OrmCoupon
        """
        return self._coupon

    @coupon.setter
    def coupon(self, coupon):
        """Sets the coupon of this OrmCouponRedemptionEdges.


        :param coupon: The coupon of this OrmCouponRedemptionEdges.  # noqa: E501
        :type coupon: OrmCoupon
        """

        self._coupon = coupon

    @property
    def organization(self):
        """Gets the organization of this OrmCouponRedemptionEdges.  # noqa: E501


        :return: The organization of this OrmCouponRedemptionEdges.  # noqa: E501
        :rtype: OrmOrganization
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """Sets the organization of this OrmCouponRedemptionEdges.


        :param organization: The organization of this OrmCouponRedemptionEdges.  # noqa: E501
        :type organization: OrmOrganization
        """

        self._organization = organization

    @property
    def user(self):
        """Gets the user of this OrmCouponRedemptionEdges.  # noqa: E501


        :return: The user of this OrmCouponRedemptionEdges.  # noqa: E501
        :rtype: OrmUser
        """
        return self._user

    @user.setter
    def user(self, user):
        """Sets the user of this OrmCouponRedemptionEdges.


        :param user: The user of this OrmCouponRedemptionEdges.  # noqa: E501
        :type user: OrmUser
        """

        self._user = user

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmCouponRedemptionEdges):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmCouponRedemptionEdges):
            return True

        return self.to_dict() != other.to_dict()



"""
Common options for serving.
"""
import sys
from typing import List
from typing import Optional

import click
import inquirer

from vessl.openapi_client import (
    ResponseModelServiceInfo,
    ModelserviceModelServiceListResponse,
    ModelserviceModelServiceRevisionListResponse,
    ResponseModelServiceRevision,
    V1Autoscaling
)
from vessl import vessl_api
from vessl.util.common import parse_time_to_ago
from vessl.util.echo import print_info, print_error
from vessl.util.exception import VesslApiException
from vessl.util.prompt import prompt_confirm, prompt_text


class RevisionOption:
    def __init__(self, number: int, description: str) -> None:
        self.number = number
        self.description = description
    
    def __str__(self):
        return self.description

def prompt_revisions_list(service_name: str) -> int:
    try:
        revisions: ModelserviceModelServiceRevisionListResponse = vessl_api.model_service_revision_list_api(
            organization_name=vessl_api.organization.name,
            model_service_name=service_name
        )
    except VesslApiException as e:
        print(f"Error occurred while fetching revision list for organization {vessl_api.organization.name}.")
    
    if len(revisions.results) == 0:
        print("No revisions found. To create revision, use 'service create' command")
        sys.exit(1)

    revisionOptions = [
        RevisionOption(number=x.number, description=f'Revision #{x.number}: {x.status} (Created {parse_time_to_ago(x.created_dt)})')
        for x in revisions.results
    ]
    choice: int = inquirer.prompt(
        [
            inquirer.List(
                "revision",
                message="Select from revision list",
                choices=revisionOptions,
            )
        ],
        raise_keyboard_interrupt=True,
    ).get("revision")
    return choice.number

def prompt_services_list() -> str:
    try:
        services: ModelserviceModelServiceListResponse = vessl_api.model_service_list_api(vessl_api.organization.name)
    except VesslApiException as e:
        print(f"There was an error while fetching service list for organization {vessl_api.organization.name}.")
    
    if len(services.results) == 0:
        print("No services found")
        sys.exit(1)
    service_names = [x.name for x in services.results]
    service_name = inquirer.prompt(
        [
            inquirer.List(
                "service",
                message="Select from service list",
                choices=service_names,
            )
        ],
        raise_keyboard_interrupt=True,
    ).get("service")
    return service_name

def build_autoscaling_configs(conf: V1Autoscaling, min:int, max:int, target: int, metric:str) -> V1Autoscaling:
    if conf is None:
        conf = V1Autoscaling(
            min=1,
            max=1,
            metric="cpu",
            target=60,
        )

    if any(x is not None for x in [min, max, target, metric]):
        if min is not None:
            conf.min = min
        if max is not None:
            conf.max = max
        if target is not None:
            conf.target = conf
        if metric is not None:
            conf.metric = metric

    return conf

def revision_callback(
    ctx: click.Context, param: click.Parameter, revision_number: Optional[int]
) -> ResponseModelServiceRevision:
    
    if vessl_api.organization is None:
            vessl_api.set_organization()
    service: ResponseModelServiceInfo = ctx.serviceObj
    
    if revision_number is None:
        revision_number = prompt_revisions_list(service_name=service.name)
    
    try:
        revision: ResponseModelServiceRevision = vessl_api.model_service_revision_read_api(
            organization_name=vessl_api.organization.name,
            model_service_name=service.name,
            revision_number=revision_number
        )
    except VesslApiException as e:
        print(f"Error occurred while reading revision #{revision_number} of service {service.name}: {e.message}")
        sys.exit(1)

    return revision
    

def service_name_callback(
    ctx: click.Context, param: click.Parameter, service_name: Optional[str]
) -> ResponseModelServiceInfo:
    if vessl_api.organization is None:
        vessl_api.set_organization()

    if service_name is None:
        service_name = prompt_services_list()

    try:
        service: ResponseModelServiceInfo = vessl_api.model_service_read_api(
            service_name, vessl_api.organization.name
        )
    except VesslApiException as e:
        print(f"Invalid service of name {service_name}: {e.message}")
        sys.exit(1)
    
    ctx.serviceObj = service

    return service

def service_name_string_callback(
    ctx: click.Context, param: click.Parameter, service_name: Optional[str]
) -> str:
    if vessl_api.organization is None:
        vessl_api.set_organization()
    if service_name is None:
        options = []
        try:
            services: ModelserviceModelServiceListResponse = vessl_api.model_service_list_api(vessl_api.organization.name)
            options = [x.name for x in services.results]
        except VesslApiException as e:
            print(f"There was an error while fetching service list for organization {vessl_api.organization.name}.")
            print()
        options.append("Create new service")
        service_name = inquirer.prompt(
            [
                inquirer.List(
                    "service",
                    message="Create revision on existing service or new service",
                    choices=options
                )
            ],
            raise_keyboard_interrupt=True,
        ).get("service")
    if service_name == "Create new service":
        service_name = prompt_text("Enter new service name")
    return service_name

def multiple_rev_nums_callback(
    ctx: click.Context, param: click.Parameter, nums: Optional[List[int]]
) -> List[int]:
    if nums is None or len(nums) == 0:
        target_revision_numbers = []
        rev = prompt_revisions_list(service_name=ctx.serviceObj.name)
        target_revision_numbers.append(rev)
        while True:
            nth_revision = prompt_confirm("Do you want to send traffic to other revisions?")
            if nth_revision:
                revision_number = prompt_revisions_list(service_name=ctx.serviceObj.name)
                target_revision_numbers.append(revision_number)
            else:
                break
        ctx.targetNums = target_revision_numbers
        return target_revision_numbers
    else:
        return nums

def traffic_weights_callback(
    ctx: click.Context, param: click.Parameter, weights: Optional[List[int]]
) -> List[int]:
    if weights is None or len(weights) == 0:
        weights = []
        target_revision_numbers = ctx.targetNums
        target_revision_numbers.sort()
        for i, num in enumerate(target_revision_numbers):
            if i == len(target_revision_numbers) -1:
                remaining_weight = 100 - sum(weights)
                print_info(f"Last revision will take rest of the traffic.({remaining_weight}%)")
                weights.append(remaining_weight)
            else:
                w = prompt_text(f"How much traffic should revision {num} receive?(in percentage)")
                if not w.isdigit():
                    print_error("Please type an integer between 0 and 100.")
                    continue
                elif int(w) > 100 or int(w) < 0:
                    print_error("Please type an integer between 0 and 100.")
                    continue
                elif sum(weights) + int(w) > 100:
                    print_error("Total weight should not exceed 100.")
                    continue
                else:
                    weights.append(int(w))
        return weights
    else:
        return weights

service_name_option = click.option(
    "--service",
    type=click.STRING,
    callback=service_name_callback,
    help="Name of service.",
)

service_name_or_new_option = click.option(
    "-s",
    "--service-name",
    type=click.STRING,
    callback=service_name_string_callback,
    help="Name of service to create this revision inside. "
    "If such service does not exist, you will be be asked on whether to create one. "
    "If service name is not given, but YAML has a name field, "
    "then that name will be used.",
)

revision_option = click.option(
    "--number",
    "-n",
    type=int,
    callback=revision_callback,
    help="Number of revision."
)

detail_option = click.option(
    "-d",
    "--detail",
    type=click.BOOL,
    is_flag=True,
    default=False,
    help="show details about each revision in service.",
)

force_option = click.option(
    "--force",
    type=click.BOOL,
    is_flag=True,
    help="force to abort the existing rollout.",
)

no_prompt_option = click.option(
    "-y",
    "--no-prompt",
    type=click.BOOL,
    is_flag=True,
    help="do not ask anything while creating revision"
)

hub_key_option = click.option(
    "--from-hub",
    type=click.STRING,
    help="Model key to be found in VESSL Hub. This will automatically create service revision from YAML file.",
)

file_path_option = click.option(
    "-f",
    "--file",
    type=click.File("r"),
    help="Path to YAML file for service revision definition.",
)

service_launch_option = click.option(
    "-l",
    "--launch",
    type=click.BOOL,
    is_flag=True,
    help="Launch after creating revision.",
)

service_activate_option = click.option(
    "-a",
    "--set-current-active",
    type=click.BOOL,
    is_flag=True,
    help="Launch and send traffic to the created revision. If another live revision exists, replace the existing revision.",
)

serverless_service_option = click.option(
    "--serverless",
    type=click.BOOL,
    is_flag=True,
    help="If the service does not exist and should be created, this flag requires "
    "the service to be serverless. If not given, it defaults to provisioned. "
    "If this parameter is given, but the named service is not in serverless mode, "
    "an error is raised.",
)

nodeport_option = click.option(
    "--port",
    type=click.INT,
    required=False,
    help=(
        "When creating a new service, the cluster that custom endpoint type is NodePort requires "
        "a port number to be specified. This is the port number that the service will be exposed on."
    ),
)

subdomain_option = click.option(
    "--subdomain",
    type=click.STRING,
    required=False,
    help=(
        "When creating a new service, the cluster that custom endpoint type is Subdomain requires "
        "a subdomain to be specified. This is the subdomain that the service will be exposed on."
    ),
)

update_gateway_option = click.option(
    "--update-gateway/--no-update-gateway",
    "-g/-G",
    is_flag=True,
    default=False,
    help="Whether to update gateway so that it points to this revision.",
)

enable_gateway_if_off_option = click.option(
    "--enable-gateway-if-off/--no-enable-gateway-if-off",
    "-e/-E",
    is_flag=True,
    default=False,
    help="When updating gateway, whether to enable the gateway if it is currently off.",
)

update_gateway_weight_option = click.option(
    "--update-gateway-weight",
    type=click.INT,
    required=False,
    help=(
        "When updating gateway, the amount of traffic that should be "
        "directed to this revision (in percentage)."
    ),
)

update_gateway_port_option = click.option(
    "--update-gateway-port",
    type=click.INT,
    required=False,
    help=(
        "When updating gateway, the port to receive the traffic; "
        "this port must be defined in serving spec."
    ),
)

format_option = click.option(
    "--format",
    type=click.STRING,
    required=True,
    default="text",
    help=("Determines output format, supports text, json"),
)

autoscaling_min_option = click.option(
    "--min",
    required=False,
    type=int,
    help="Number of min replicas."
)

autoscaling_max_option = click.option(
    "--max",
    required=False,
    type=int,
    help="Number of max replicas."
)

autoscaling_target_option = click.option(
    "--target",
    required=False,
    type=int,
    help="Target resource utilization."
)

autoscaling_metric_option = click.option(
    "--metric",
    required=False,
    type=str,
    help="Metric for autoscaling. "
    "Supported autoscaling metrics: 'cpu', 'nvidia.com/gpu', 'memory'"
)

split_traffic_revs_option = click.option(
    "--number",
    "-n",
    required=False,
    type=int,
    multiple=True,
    callback=multiple_rev_nums_callback,
    help="Number of revisions to launch."
)

split_traffic_weights_option = click.option(
    "--weight",
    "-w",
    required=False,
    type=int,
    multiple=True,
    callback=traffic_weights_callback,
    help="Number of traffic weight for the revision",
)


# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import copy
import logging
import multiprocessing
import sys
import urllib3

import six
from six.moves import http_client as httplib
from vessl.openapi_client.exceptions import ApiValueError


JSON_SCHEMA_VALIDATION_KEYWORDS = {
    'multipleOf', 'maximum', 'exclusiveMaximum',
    'minimum', 'exclusiveMinimum', 'maxLength',
    'minLength', 'pattern', 'maxItems', 'minItems'
}

class Configuration(object):
    """NOTE: This class is auto generated by OpenAPI Generator

    Ref: https://openapi-generator.tech
    Do not edit the class manually.

    :param host: Base url
    :param api_key: Dict to store API key(s).
      Each entry in the dict specifies an API key.
      The dict key is the name of the security scheme in the OAS specification.
      The dict value is the API key secret.
    :param api_key_prefix: Dict to store API prefix (e.g. Bearer)
      The dict key is the name of the security scheme in the OAS specification.
      The dict value is an API key prefix when generating the auth data.
    :param username: Username for HTTP basic authentication
    :param password: Password for HTTP basic authentication
    :param discard_unknown_keys: Boolean value indicating whether to discard
      unknown properties. A server may send a response that includes additional
      properties that are not known by the client in the following scenarios:
      1. The OpenAPI document is incomplete, i.e. it does not match the server
         implementation.
      2. The client was generated using an older version of the OpenAPI document
         and the server has been upgraded since then.
      If a schema in the OpenAPI document defines the additionalProperties attribute,
      then all undeclared properties received by the server are injected into the
      additional properties map. In that case, there are undeclared properties, and
      nothing to discard.
    :param disabled_client_side_validations (string): Comma-separated list of
      JSON schema validation keywords to disable JSON schema structural validation
      rules. The following keywords may be specified: multipleOf, maximum,
      exclusiveMaximum, minimum, exclusiveMinimum, maxLength, minLength, pattern,
      maxItems, minItems.
      By default, the validation is performed for data generated locally by the client
      and data received from the server, independent of any validation performed by
      the server side. If the input data does not satisfy the JSON schema validation
      rules specified in the OpenAPI document, an exception is raised.
      If disabled_client_side_validations is set, structural validation is
      disabled. This can be useful to troubleshoot data validation problem, such as
      when the OpenAPI document validation rules do not match the actual API data
      received by the server.
    :param server_index: Index to servers configuration.
    :param server_variables: Mapping with string values to replace variables in
      templated server configuration. The validation of enums is performed for
      variables with defined enum values before.
    :param server_operation_index: Mapping from operation ID to an index to server
      configuration.
    :param server_operation_variables: Mapping from operation ID to a mapping with
      string values to replace variables in templated server configuration.
      The validation of enums is performed for variables with defined enum values before.
    :param ssl_ca_cert: str - the path to a file of concatenated CA certificates
      in PEM format

    """

    _default = None

    def __init__(self, host=None,
                 api_key=None, api_key_prefix=None,
                 username=None, password=None,
                 discard_unknown_keys=False,
                 disabled_client_side_validations="",
                 server_index=None, server_variables=None,
                 server_operation_index=None, server_operation_variables=None,
                 ssl_ca_cert=None,
                 ):
        """Constructor
        """
        self._base_path = "http://localhost" if host is None else host
        """Default Base url
        """
        self.server_index = 0 if server_index is None and host is None else server_index
        self.server_operation_index = server_operation_index or {}
        """Default server index
        """
        self.server_variables = server_variables or {}
        self.server_operation_variables = server_operation_variables or {}
        """Default server variables
        """
        self.temp_folder_path = None
        """Temp file folder for downloading files
        """
        # Authentication Settings
        self.api_key = {}
        if api_key:
            self.api_key = api_key
        """dict to store API key(s)
        """
        self.api_key_prefix = {}
        if api_key_prefix:
            self.api_key_prefix = api_key_prefix
        """dict to store API prefix (e.g. Bearer)
        """
        self.refresh_api_key_hook = None
        """function hook to refresh API key if expired
        """
        self.username = username
        """Username for HTTP basic authentication
        """
        self.password = password
        """Password for HTTP basic authentication
        """
        self.discard_unknown_keys = discard_unknown_keys
        self.disabled_client_side_validations = disabled_client_side_validations
        self.logger = {}
        """Logging Settings
        """
        self.logger["package_logger"] = logging.getLogger("vessl.openapi_client")
        self.logger["urllib3_logger"] = logging.getLogger("urllib3")
        self.logger_format = '%(asctime)s %(levelname)s %(message)s'
        """Log format
        """
        self.logger_stream_handler = None
        """Log stream handler
        """
        self.logger_file_handler = None
        """Log file handler
        """
        self.logger_file = None
        """Debug file location
        """
        self.debug = False
        """Debug switch
        """

        self.verify_ssl = True
        """SSL/TLS verification
           Set this to false to skip verifying SSL certificate when calling API
           from https server.
        """
        self.ssl_ca_cert = ssl_ca_cert
        """Set this to customize the certificate file to verify the peer.
        """
        self.cert_file = None
        """client certificate file
        """
        self.key_file = None
        """client key file
        """
        self.assert_hostname = None
        """Set this to True/False to enable/disable SSL hostname verification.
        """

        self.connection_pool_maxsize = multiprocessing.cpu_count() * 5
        """urllib3 connection pool's maximum number of connections saved
           per pool. urllib3 uses 1 connection as default value, but this is
           not the best value when you are making a lot of possibly parallel
           requests to the same host, which is often the case here.
           cpu_count * 5 is used as default value to increase performance.
        """

        self.proxy = None
        """Proxy URL
        """
        self.proxy_headers = None
        """Proxy headers
        """
        self.safe_chars_for_path_param = ''
        """Safe chars for path_param
        """
        self.retries = None
        """Adding retries to override urllib3 default value 3
        """
        # Enable client side validation
        self.client_side_validation = True

        self.socket_options = None
        """Options to pass down to the underlying urllib3 socket
        """

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k not in ('logger', 'logger_file_handler'):
                setattr(result, k, copy.deepcopy(v, memo))
        # shallow copy of loggers
        result.logger = copy.copy(self.logger)
        # use setters to configure loggers
        result.logger_file = self.logger_file
        result.debug = self.debug
        return result

    def __setattr__(self, name, value):
        object.__setattr__(self, name, value)
        if name == 'disabled_client_side_validations':
            s = set(filter(None, value.split(',')))
            for v in s:
                if v not in JSON_SCHEMA_VALIDATION_KEYWORDS:
                    raise ApiValueError(
                        "Invalid keyword: '{0}''".format(v))
            self._disabled_client_side_validations = s

    @classmethod
    def set_default(cls, default):
        """Set default instance of configuration.

        It stores default configuration, which can be
        returned by get_default_copy method.

        :param default: object of Configuration
        """
        cls._default = copy.deepcopy(default)

    @classmethod
    def get_default_copy(cls):
        """Return new instance of configuration.

        This method returns newly created, based on default constructor,
        object of Configuration class or returns a copy of default
        configuration passed by the set_default method.

        :return: The configuration object.
        """
        if cls._default is not None:
            return copy.deepcopy(cls._default)
        return Configuration()

    @property
    def logger_file(self):
        """The logger file.

        If the logger_file is None, then add stream handler and remove file
        handler. Otherwise, add file handler and remove stream handler.

        :param value: The logger_file path.
        :type: str
        """
        return self.__logger_file

    @logger_file.setter
    def logger_file(self, value):
        """The logger file.

        If the logger_file is None, then add stream handler and remove file
        handler. Otherwise, add file handler and remove stream handler.

        :param value: The logger_file path.
        :type: str
        """
        self.__logger_file = value
        if self.__logger_file:
            # If set logging file,
            # then add file handler and remove stream handler.
            self.logger_file_handler = logging.FileHandler(self.__logger_file)
            self.logger_file_handler.setFormatter(self.logger_formatter)
            for _, logger in six.iteritems(self.logger):
                logger.addHandler(self.logger_file_handler)

    @property
    def debug(self):
        """Debug status

        :param value: The debug status, True or False.
        :type: bool
        """
        return self.__debug

    @debug.setter
    def debug(self, value):
        """Debug status

        :param value: The debug status, True or False.
        :type: bool
        """
        self.__debug = value
        if self.__debug:
            # if debug status is True, turn on debug logging
            for _, logger in six.iteritems(self.logger):
                logger.setLevel(logging.DEBUG)
            # turn on httplib debug
            httplib.HTTPConnection.debuglevel = 1
        else:
            # if debug status is False, turn off debug logging,
            # setting log level to default `logging.WARNING`
            for _, logger in six.iteritems(self.logger):
                logger.setLevel(logging.WARNING)
            # turn off httplib debug
            httplib.HTTPConnection.debuglevel = 0

    @property
    def logger_format(self):
        """The logger format.

        The logger_formatter will be updated when sets logger_format.

        :param value: The format string.
        :type: str
        """
        return self.__logger_format

    @logger_format.setter
    def logger_format(self, value):
        """The logger format.

        The logger_formatter will be updated when sets logger_format.

        :param value: The format string.
        :type: str
        """
        self.__logger_format = value
        self.logger_formatter = logging.Formatter(self.__logger_format)

    def get_api_key_with_prefix(self, identifier, alias=None):
        """Gets API key (with prefix if set).

        :param identifier: The identifier of apiKey.
        :param alias: The alternative identifier of apiKey.
        :return: The token for api key authentication.
        """
        if self.refresh_api_key_hook is not None:
            self.refresh_api_key_hook(self)
        key = self.api_key.get(identifier, self.api_key.get(alias) if alias is not None else None)
        if key:
            prefix = self.api_key_prefix.get(identifier)
            if prefix:
                return "%s %s" % (prefix, key)
            else:
                return key

    def get_basic_auth_token(self):
        """Gets HTTP basic authentication header (string).

        :return: The token for basic HTTP authentication.
        """
        username = ""
        if self.username is not None:
            username = self.username
        password = ""
        if self.password is not None:
            password = self.password
        return urllib3.util.make_headers(
            basic_auth=username + ':' + password
        ).get('authorization')

    def auth_settings(self):
        """Gets Auth Settings dict for api client.

        :return: The Auth Settings information dict.
        """
        auth = {}
        return auth

    def to_debug_report(self):
        """Gets the essential information for debugging.

        :return: The report for debugging.
        """
        return "Python SDK Debug Report:\n"\
               "OS: {env}\n"\
               "Python Version: {pyversion}\n"\
               "Version of the API: 1.0.0\n"\
               "SDK Package Version: 1.0.0".\
               format(env=sys.platform, pyversion=sys.version)

    def get_host_settings(self):
        """Gets an array of host settings

        :return: An array of host settings
        """
        return [
            {
                'url': "",
                'description': "No description provided",
            }
        ]

    def get_host_from_settings(self, index, variables=None, servers=None):
        """Gets host URL based on the index and variables
        :param index: array index of the host settings
        :param variables: hash of variable and the corresponding value
        :param servers: an array of host settings or None
        :return: URL based on host settings
        """
        if index is None:
            return self._base_path

        variables = {} if variables is None else variables
        servers = self.get_host_settings() if servers is None else servers

        try:
            server = servers[index]
        except IndexError:
            raise ValueError(
                "Invalid index {0} when selecting the host settings. "
                "Must be less than {1}".format(index, len(servers)))

        url = server['url']

        # go through variables and replace placeholders
        for variable_name, variable in server.get('variables', {}).items():
            used_value = variables.get(
                variable_name, variable['default_value'])

            if 'enum_values' in variable \
                    and used_value not in variable['enum_values']:
                raise ValueError(
                    "The variable `{0}` in the host URL has invalid value "
                    "{1}. Must be {2}.".format(
                        variable_name, variables[variable_name],
                        variable['enum_values']))

            url = url.replace("{" + variable_name + "}", used_value)

        return url

    @property
    def host(self):
        """Return generated host."""
        return self.get_host_from_settings(self.server_index, variables=self.server_variables)

    @host.setter
    def host(self, value):
        """Fix base path."""
        self._base_path = value
        self.server_index = None



import io
from typing import Dict, List, Optional, Union

import torch
from diffusers import DiffusionPipeline
from PIL import Image
from pydantic import BaseModel

from vessl.service import RunnerBase

MODEL_NAME_OR_PATH = "{MODEL_NAME_OR_PATH}"


class InputType(BaseModel):
    prompt: Union[str, List[str]]
    height: Optional[int] = None
    width: Optional[int] = None
    num_inference_steps: Optional[int] = 50
    guidance_scale: Optional[float] = 7.5
    negative_prompt: Optional[Union[str, List[str]]] = None


class OutputType(BaseModel):
    result: bytes


def get_device() -> torch.device:
    if torch.cuda.is_available():
        return torch.device("cuda")
    if torch.backends.mps.is_available():
        return torch.device("mps")
    return torch.device("cpu")


class HfDiffuserRunner(RunnerBase[InputType, OutputType]):
    @staticmethod
    def load_model(
        props: Union[Dict[str, str], None], artifacts: Dict[str, str]
    ) -> DiffusionPipeline:
        pipeline = DiffusionPipeline.from_pretrained(
            MODEL_NAME_OR_PATH, torch_dtype=torch.float16, variant="fp16"
        )

        device = get_device()
        pipeline = pipeline.to(device)

        return pipeline

    @staticmethod
    def preprocess_data(data: InputType) -> InputType:
        return data

    @staticmethod
    def predict(model: DiffusionPipeline, data: InputType) -> Image.Image:
        inputs = data.model_dump()
        outputs = model(**inputs)
        return outputs.images[0]

    @staticmethod
    def postprocess_data(data: Image.Image):
        fp = io.BytesIO()
        data.save(fp, format="PNG")
        byte_array = fp.getvalue()

        return {"result": byte_array}


if __name__ == "__main__":
    inputs = InputType(prompt="white cat")

    model = HfDiffuserRunner.load_model(None, None)
    data = HfDiffuserRunner.preprocess_data(inputs)
    result = HfDiffuserRunner.predict(model, data)
    output = HfDiffuserRunner.postprocess_data(result)



def get_hftransformers_pattern():
    return '''
from dataclasses import dataclass
from typing import Dict, Optional, Tuple, Union

import torch
from pydantic import BaseModel
from transformers import (
    {model_type},
    {tokenizer_type},
    PreTrainedModel,
    PreTrainedTokenizer,
)

from vessl.service import RunnerBase


@dataclass
class HfLanguageModel:
    tokenizer: PreTrainedTokenizer
    model: PreTrainedModel


class InputType(BaseModel):
    text: str
    max_length: Optional[int] = 20
    max_new_tokens: Optional[int] = None
    do_sample: Optional[bool] = False
    temperature: Optional[float] = 1.0
    top_k: Optional[int] = 50
    top_p: Optional[float] = 1.0
    num_beams: Optional[int] = 1
    num_beam_groups: Optional[int] = 1


class OutputType(BaseModel):
    result: str


def get_device() -> torch.device:
    if torch.cuda.is_available():
        return torch.device("cuda")
    if torch.backends.mps.is_available():
        return torch.device("mps")
    return torch.device("cpu")


class HfTransformerRunner(RunnerBase[InputType, OutputType]):
    @staticmethod
    def load_model(
        props: Union[Dict[str, str], None], artifacts: Dict[str, str]
    ) -> Tuple[PreTrainedModel, PreTrainedTokenizer]:
        # TODO: transformer has a lot of models including causal LM
        model = {model_type}.from_pretrained({model_name_or_path})
        tokenizer = {tokenizer_type}.from_pretrained({model_name_or_path})

        device = get_device()
        model = model.to(device)

        return HfLanguageModel(tokenizer, model)

    @staticmethod
    def preprocess_data(data: InputType) -> str:
        return data

    @staticmethod
    def predict(model: HfLanguageModel, data: InputType) -> str:
        tokenizer = model.tokenizer
        model = model.model

        inputs = tokenizer(data.text, return_tensors="pt").to(model.device)
        configs = data.model_dump(exclude={{"text"}})
        outputs = model.generate(**inputs, **configs)
        output_text = tokenizer.decode(outputs[0])

        return output_text

    @staticmethod
    def postprocess_data(data: str):
        return OutputType(result=data)


if __name__ == "__main__":
    inputs = InputType(text="Hello!")

    model = HfTransformerRunner.load_model(None, None)
    data = HfTransformerRunner.preprocess_data(inputs)
    result = HfTransformerRunner.predict(model, data)
    output = HfTransformerRunner.postprocess_data(result)
'''

def get_hfdiffusers_pattern():
    return '''
from dataclasses import dataclass
from typing import Dict, Optional, Tuple, Union
from PIL import Image

import torch
from pydantic import BaseModel
from transformers import (
    {pipeline_type},
)

from vessl.service import RunnerBase


@dataclass
class HfDiffuserPipeline:
    pipeline: {pipeline_type}


class InputType(BaseModel):
    prompt: str


class OutputType(BaseModel):
    result: Image


def get_device() -> torch.device:
    if torch.cuda.is_available():
        return torch.device("cuda")
    if torch.backends.mps.is_available():
        return torch.device("mps")
    return torch.device("cpu")


class HfDiffuserRunner(RunnerBase[InputType, OutputType]):
    @staticmethod
    def load_model(
        props: Union[Dict[str, str], None], artifacts: Dict[str, str]
    ) -> {pipeline_type}:
        pipeline = {pipeline_type}.from_pretrained({model_name_or_path}, torch_dtype=torch.float16, use_safetensors=True)

        device = get_device()
        pipeline = pipeline.to(device)

        return HfDiffuserPipeline(pipeline)

    @staticmethod
    def preprocess_data(data: InputType) -> str:
        return data

    @staticmethod
    def predict(pipeline: HfDiffuserPipeline, data: InputType) -> Image:
        pipeline = pipeline.pipeline
        generator = torch.Generator("cuda").manual_seed(0)
        output = pipeline(data.prompt, generator=generator).images[0]

        return output

    @staticmethod
    def postprocess_data(data: Image):
        return OutputType(result=data)


if __name__ == "__main__":
    inputs = InputType(text="Hello!")

    model = HfTransformerRunner.load_model(None, None)
    data = HfTransformerRunner.preprocess_data(inputs)
    result = HfTransformerRunner.predict(model, data)
    output = HfTransformerRunner.postprocess_data(result)
'''


from typing import Literal, Optional

from vessl.openapi_client import (
    ProtoLLMModelConnection,
    ProtoLLMUserGroup,
    ProtoLLMVectorDBConnConfig,
    ProtoRevealedSecret,
    V1LLMPermissionConfig,
)
from vessl.llm import connection_auth


class ModelSecret(object):
    kind: Literal["openai", "aws"]


class ModelSecretAWS(ModelSecret):
    kind: Literal["aws"]
    access_key: str
    secret_key: str

    def __init__(self, access_key: str, secret_key: str):
        self.kind = "aws"
        self.access_key = access_key
        self.secret_key = secret_key


class ModelSecretOpenAI(ModelSecret):
    kind: Literal["openai"]
    api_key: str

    def __init__(self, api_key: str):
        self.kind = "openai"
        self.api_key = api_key


class ModelConnection(object):
    id: int
    name: str
    model_type: Literal["embedding", "llm"]
    url: str = ""
    api_key: str | None = None
    model_secret: ModelSecret | None = None

    def __init__(self, model_connection: ProtoLLMModelConnection):
        self.id = model_connection.id
        self.name = model_connection.name
        if model_connection.type == "embedding":
            self.model_type = "embedding"
        elif model_connection.type == "llm":
            self.model_type = "llm"
        self.url = model_connection.url
        self.api_key = model_connection.api_key
        model_secret = model_connection.revealed_secret
        if model_secret is not None:
            if model_secret.kind == 'openai':
                self.model_secret = ModelSecretOpenAI(model_secret.openai.api_key.openai_api_key)
            elif model_secret.kind == 'aws':
                if model_secret.aws.credential_type == 'access_key':
                    access_key = model_secret.aws.access_key
                    self.model_secret = ModelSecretAWS(access_key.access_key_id, access_key.secret_access_key)


class VectorDBConnection(object):
    id: int
    name: str
    vector_db_type: Literal["chroma"]
    url: str
    collection_name: str
    authentication: connection_auth.ConnectionAuth

    def __init__(self, vector_db: ProtoLLMVectorDBConnConfig, collection_name: str):
        self.id = vector_db.id
        self.name = vector_db.name
        if vector_db.vdb_type == "chroma":
            self.vector_db_type = "chroma"
        self.url = vector_db.url
        self.collection_name = collection_name
        self.authentication = connection_auth.ConnectionAuth(vector_db.authentication)


class PermissionConfig(object):
    user_group_ids: Optional[list[int]]

    def __init__(self, permission_config: V1LLMPermissionConfig):
        self.user_group_ids = permission_config.user_group_ids


class UserGroup(object):
    id: int
    name: str
    description: Optional[str]
    created_by_id: int
    created_dt: float
    organization_id: int
    updated_dt: float

    def __init__(self, user_group: ProtoLLMUserGroup):
        self.id = user_group.id
        self.name = user_group.name
        self.description = user_group.description
        self.created_by_id = user_group.created_by_id
        self.created_dt = user_group.created_dt
        self.organization_id = user_group.organization_id
        self.updated_dt = user_group.updated_dt


class WorkflowConnectionConfiguration:
    embedding_model: Optional[ModelConnection]
    llm_model: Optional[ModelConnection]
    vector_db: Optional[VectorDBConnection]
    permission: Optional[PermissionConfig]

    def __init__(self, model: Optional[ProtoLLMModelConnection], model_secret: Optional[ProtoRevealedSecret],
                 vector_db: Optional[ProtoLLMVectorDBConnConfig],
                 permission: Optional[V1LLMPermissionConfig], vector_db_collection_name: Optional[str]):
        if model:
            if model.type == "embedding":
                self.embedding_model = ModelConnection(model, model_secret)
            elif model.type == "llm":
                self.llm_model = ModelConnection(model, model_secret)
        if vector_db:
            self.vector_db = VectorDBConnection(vector_db, vector_db_collection_name)
        if permission:
            self.permission = PermissionConfig(permission)



import os

import click
import inquirer

from vessl.cli._base import VesslGroup, vessl_option
from vessl.cli._util import (
  print_table,
  truncate_datetime,
)
from vessl.util.prompt import prompt_choices
from vessl.util.echo import print_success, print_info
from vessl.ssh_key import create_ssh_key, delete_ssh_key, list_ssh_keys
from vessl.util.common import parse_time_to_ago
from vessl.util.constant import SSH_PUBLIC_KEY_PATH
from vessl.util.ssh import set_default_ssh_key


def ssh_key_id_callback(ctx: click.Context, param: click.Parameter, value: id) -> id:
    ssh_keys = list_ssh_keys()
    return prompt_choices(
        "SSH key",
        [
            (
                f"{x.name} / {x.fingerprint} (created {parse_time_to_ago(x.created_dt)})",
                x.id,
            )
            for x in ssh_keys
        ],
    )


def ssh_public_key_value_callback(ctx: click.Context, param: click.Option, key_path: str) -> str:
    from sshpubkeys import InvalidKeyError, SSHKey

    if not key_path:
        key_path = inquirer.prompt(
            [
                inquirer.Text(
                    "question",
                    message="SSH public key path",
                    default=SSH_PUBLIC_KEY_PATH,
                )
            ],
            raise_keyboard_interrupt=True,
        ).get("question")

    try:
        with open(key_path, "r") as f:
            ssh_public_key_value = f.read()
    except FileNotFoundError:
        raise click.BadParameter("Key file not found")

    ssh = SSHKey(ssh_public_key_value, strict=True)
    try:
        ssh.parse()
    except InvalidKeyError as e:
        raise click.BadParameter(f"Invalid key: {e}")
    except NotImplementedError as e:
        raise click.BadParameter(f"Invalid key type: {e}")

    ctx.obj["ssh_key_name"] = ssh.comment
    ctx.obj["ssh_key_filename"] = os.path.basename(key_path)
    ctx.obj["ssh_key_filepath"] = key_path

    return ssh_public_key_value


def ssh_key_name_callback(ctx: click.Context, param: click.Option, name: str) -> str:
    if name:
        return name

    return inquirer.prompt(
        [
            inquirer.Text(
                "question",
                message="SSH public key name",
                default=ctx.obj.get("ssh_key_name"),
            )
        ],
        raise_keyboard_interrupt=True,
    ).get("question")


@click.command(name="ssh-key", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
def list():
    ssh_keys = list_ssh_keys()
    print_table(
        ssh_keys,
        ["Name", "Fingerprint", "Created"],
        lambda x: [x.name, x.fingerprint, truncate_datetime(x.created_dt)],
    )


@cli.vessl_command()
@vessl_option(
    "-p",
    "--path",
    "ssh_public_key_value",
    type=click.Path(exists=True),
    help="Path to SSH public key.",
    callback=ssh_public_key_value_callback,
)
@vessl_option(
    "--name",
    type=click.STRING,
    callback=ssh_key_name_callback,
)
@click.pass_context
def add(ctx: click.Context, name: str, ssh_public_key_value: str):
    ssh_key = create_ssh_key(
        key_name=name,
        key_path=ctx.obj["ssh_key_filename"],
        ssh_public_key_value=ssh_public_key_value,
    )
    set_default_ssh_key(ctx.obj["ssh_key_filepath"])

    if ssh_key.name != name:
        print_info(f"Identical SSH Key '{ssh_key.name}' already exists, skipping creation.")
    else:
        print_success(f"SSH Key '{ssh_key.name}' created.")


@cli.vessl_command()
@vessl_option(
    "--id",
    type=click.INT,
    hidden=True,
    callback=ssh_key_id_callback,
)
def delete(id: int):
    delete_ssh_key(key_id=id)
    print_success("Successfully deleted.")



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class SamlIDPSSODescriptor(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'artifact_resolution_services': 'list[SamlEndpoint]',
        'assertion_id_request_services': 'list[SamlEndpoint]',
        'attribute_profiles': 'list[str]',
        'attributes': 'list[SamlAttribute]',
        'cache_duration': 'str',
        'contact_people': 'list[SamlContactPerson]',
        'error_url': 'str',
        'id': 'str',
        'key_descriptors': 'list[SamlKeyDescriptor]',
        'manage_name_id_services': 'list[SamlEndpoint]',
        'name_id_formats': 'list[str]',
        'name_id_mapping_services': 'list[SamlEndpoint]',
        'organization': 'SamlOrganization',
        'protocol_support_enumeration': 'str',
        'signature': 'EtreeElement',
        'single_logout_services': 'list[SamlEndpoint]',
        'single_sign_on_services': 'list[SamlEndpoint]',
        'valid_until': 'datetime',
        'want_authn_requests_signed': 'bool',
        'xml_name': 'XmlName'
    }

    attribute_map = {
        'artifact_resolution_services': 'ArtifactResolutionServices',
        'assertion_id_request_services': 'AssertionIDRequestServices',
        'attribute_profiles': 'AttributeProfiles',
        'attributes': 'Attributes',
        'cache_duration': 'CacheDuration',
        'contact_people': 'ContactPeople',
        'error_url': 'ErrorURL',
        'id': 'ID',
        'key_descriptors': 'KeyDescriptors',
        'manage_name_id_services': 'ManageNameIDServices',
        'name_id_formats': 'NameIDFormats',
        'name_id_mapping_services': 'NameIDMappingServices',
        'organization': 'Organization',
        'protocol_support_enumeration': 'ProtocolSupportEnumeration',
        'signature': 'Signature',
        'single_logout_services': 'SingleLogoutServices',
        'single_sign_on_services': 'SingleSignOnServices',
        'valid_until': 'ValidUntil',
        'want_authn_requests_signed': 'WantAuthnRequestsSigned',
        'xml_name': 'XMLName'
    }

    def __init__(self, artifact_resolution_services=None, assertion_id_request_services=None, attribute_profiles=None, attributes=None, cache_duration=None, contact_people=None, error_url=None, id=None, key_descriptors=None, manage_name_id_services=None, name_id_formats=None, name_id_mapping_services=None, organization=None, protocol_support_enumeration=None, signature=None, single_logout_services=None, single_sign_on_services=None, valid_until=None, want_authn_requests_signed=None, xml_name=None, local_vars_configuration=None):  # noqa: E501
        """SamlIDPSSODescriptor - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._artifact_resolution_services = None
        self._assertion_id_request_services = None
        self._attribute_profiles = None
        self._attributes = None
        self._cache_duration = None
        self._contact_people = None
        self._error_url = None
        self._id = None
        self._key_descriptors = None
        self._manage_name_id_services = None
        self._name_id_formats = None
        self._name_id_mapping_services = None
        self._organization = None
        self._protocol_support_enumeration = None
        self._signature = None
        self._single_logout_services = None
        self._single_sign_on_services = None
        self._valid_until = None
        self._want_authn_requests_signed = None
        self._xml_name = None
        self.discriminator = None

        if artifact_resolution_services is not None:
            self.artifact_resolution_services = artifact_resolution_services
        if assertion_id_request_services is not None:
            self.assertion_id_request_services = assertion_id_request_services
        if attribute_profiles is not None:
            self.attribute_profiles = attribute_profiles
        if attributes is not None:
            self.attributes = attributes
        if cache_duration is not None:
            self.cache_duration = cache_duration
        if contact_people is not None:
            self.contact_people = contact_people
        if error_url is not None:
            self.error_url = error_url
        if id is not None:
            self.id = id
        if key_descriptors is not None:
            self.key_descriptors = key_descriptors
        if manage_name_id_services is not None:
            self.manage_name_id_services = manage_name_id_services
        if name_id_formats is not None:
            self.name_id_formats = name_id_formats
        if name_id_mapping_services is not None:
            self.name_id_mapping_services = name_id_mapping_services
        if organization is not None:
            self.organization = organization
        if protocol_support_enumeration is not None:
            self.protocol_support_enumeration = protocol_support_enumeration
        if signature is not None:
            self.signature = signature
        if single_logout_services is not None:
            self.single_logout_services = single_logout_services
        if single_sign_on_services is not None:
            self.single_sign_on_services = single_sign_on_services
        self.valid_until = valid_until
        self.want_authn_requests_signed = want_authn_requests_signed
        if xml_name is not None:
            self.xml_name = xml_name

    @property
    def artifact_resolution_services(self):
        """Gets the artifact_resolution_services of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The artifact_resolution_services of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[SamlEndpoint]
        """
        return self._artifact_resolution_services

    @artifact_resolution_services.setter
    def artifact_resolution_services(self, artifact_resolution_services):
        """Sets the artifact_resolution_services of this SamlIDPSSODescriptor.


        :param artifact_resolution_services: The artifact_resolution_services of this SamlIDPSSODescriptor.  # noqa: E501
        :type artifact_resolution_services: list[SamlEndpoint]
        """

        self._artifact_resolution_services = artifact_resolution_services

    @property
    def assertion_id_request_services(self):
        """Gets the assertion_id_request_services of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The assertion_id_request_services of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[SamlEndpoint]
        """
        return self._assertion_id_request_services

    @assertion_id_request_services.setter
    def assertion_id_request_services(self, assertion_id_request_services):
        """Sets the assertion_id_request_services of this SamlIDPSSODescriptor.


        :param assertion_id_request_services: The assertion_id_request_services of this SamlIDPSSODescriptor.  # noqa: E501
        :type assertion_id_request_services: list[SamlEndpoint]
        """

        self._assertion_id_request_services = assertion_id_request_services

    @property
    def attribute_profiles(self):
        """Gets the attribute_profiles of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The attribute_profiles of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[str]
        """
        return self._attribute_profiles

    @attribute_profiles.setter
    def attribute_profiles(self, attribute_profiles):
        """Sets the attribute_profiles of this SamlIDPSSODescriptor.


        :param attribute_profiles: The attribute_profiles of this SamlIDPSSODescriptor.  # noqa: E501
        :type attribute_profiles: list[str]
        """

        self._attribute_profiles = attribute_profiles

    @property
    def attributes(self):
        """Gets the attributes of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The attributes of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[SamlAttribute]
        """
        return self._attributes

    @attributes.setter
    def attributes(self, attributes):
        """Sets the attributes of this SamlIDPSSODescriptor.


        :param attributes: The attributes of this SamlIDPSSODescriptor.  # noqa: E501
        :type attributes: list[SamlAttribute]
        """

        self._attributes = attributes

    @property
    def cache_duration(self):
        """Gets the cache_duration of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The cache_duration of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: str
        """
        return self._cache_duration

    @cache_duration.setter
    def cache_duration(self, cache_duration):
        """Sets the cache_duration of this SamlIDPSSODescriptor.


        :param cache_duration: The cache_duration of this SamlIDPSSODescriptor.  # noqa: E501
        :type cache_duration: str
        """

        self._cache_duration = cache_duration

    @property
    def contact_people(self):
        """Gets the contact_people of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The contact_people of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[SamlContactPerson]
        """
        return self._contact_people

    @contact_people.setter
    def contact_people(self, contact_people):
        """Sets the contact_people of this SamlIDPSSODescriptor.


        :param contact_people: The contact_people of this SamlIDPSSODescriptor.  # noqa: E501
        :type contact_people: list[SamlContactPerson]
        """

        self._contact_people = contact_people

    @property
    def error_url(self):
        """Gets the error_url of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The error_url of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: str
        """
        return self._error_url

    @error_url.setter
    def error_url(self, error_url):
        """Sets the error_url of this SamlIDPSSODescriptor.


        :param error_url: The error_url of this SamlIDPSSODescriptor.  # noqa: E501
        :type error_url: str
        """

        self._error_url = error_url

    @property
    def id(self):
        """Gets the id of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The id of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this SamlIDPSSODescriptor.


        :param id: The id of this SamlIDPSSODescriptor.  # noqa: E501
        :type id: str
        """

        self._id = id

    @property
    def key_descriptors(self):
        """Gets the key_descriptors of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The key_descriptors of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[SamlKeyDescriptor]
        """
        return self._key_descriptors

    @key_descriptors.setter
    def key_descriptors(self, key_descriptors):
        """Sets the key_descriptors of this SamlIDPSSODescriptor.


        :param key_descriptors: The key_descriptors of this SamlIDPSSODescriptor.  # noqa: E501
        :type key_descriptors: list[SamlKeyDescriptor]
        """

        self._key_descriptors = key_descriptors

    @property
    def manage_name_id_services(self):
        """Gets the manage_name_id_services of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The manage_name_id_services of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[SamlEndpoint]
        """
        return self._manage_name_id_services

    @manage_name_id_services.setter
    def manage_name_id_services(self, manage_name_id_services):
        """Sets the manage_name_id_services of this SamlIDPSSODescriptor.


        :param manage_name_id_services: The manage_name_id_services of this SamlIDPSSODescriptor.  # noqa: E501
        :type manage_name_id_services: list[SamlEndpoint]
        """

        self._manage_name_id_services = manage_name_id_services

    @property
    def name_id_formats(self):
        """Gets the name_id_formats of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The name_id_formats of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[str]
        """
        return self._name_id_formats

    @name_id_formats.setter
    def name_id_formats(self, name_id_formats):
        """Sets the name_id_formats of this SamlIDPSSODescriptor.


        :param name_id_formats: The name_id_formats of this SamlIDPSSODescriptor.  # noqa: E501
        :type name_id_formats: list[str]
        """

        self._name_id_formats = name_id_formats

    @property
    def name_id_mapping_services(self):
        """Gets the name_id_mapping_services of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The name_id_mapping_services of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[SamlEndpoint]
        """
        return self._name_id_mapping_services

    @name_id_mapping_services.setter
    def name_id_mapping_services(self, name_id_mapping_services):
        """Sets the name_id_mapping_services of this SamlIDPSSODescriptor.


        :param name_id_mapping_services: The name_id_mapping_services of this SamlIDPSSODescriptor.  # noqa: E501
        :type name_id_mapping_services: list[SamlEndpoint]
        """

        self._name_id_mapping_services = name_id_mapping_services

    @property
    def organization(self):
        """Gets the organization of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The organization of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: SamlOrganization
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """Sets the organization of this SamlIDPSSODescriptor.


        :param organization: The organization of this SamlIDPSSODescriptor.  # noqa: E501
        :type organization: SamlOrganization
        """

        self._organization = organization

    @property
    def protocol_support_enumeration(self):
        """Gets the protocol_support_enumeration of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The protocol_support_enumeration of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: str
        """
        return self._protocol_support_enumeration

    @protocol_support_enumeration.setter
    def protocol_support_enumeration(self, protocol_support_enumeration):
        """Sets the protocol_support_enumeration of this SamlIDPSSODescriptor.


        :param protocol_support_enumeration: The protocol_support_enumeration of this SamlIDPSSODescriptor.  # noqa: E501
        :type protocol_support_enumeration: str
        """

        self._protocol_support_enumeration = protocol_support_enumeration

    @property
    def signature(self):
        """Gets the signature of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The signature of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: EtreeElement
        """
        return self._signature

    @signature.setter
    def signature(self, signature):
        """Sets the signature of this SamlIDPSSODescriptor.


        :param signature: The signature of this SamlIDPSSODescriptor.  # noqa: E501
        :type signature: EtreeElement
        """

        self._signature = signature

    @property
    def single_logout_services(self):
        """Gets the single_logout_services of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The single_logout_services of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[SamlEndpoint]
        """
        return self._single_logout_services

    @single_logout_services.setter
    def single_logout_services(self, single_logout_services):
        """Sets the single_logout_services of this SamlIDPSSODescriptor.


        :param single_logout_services: The single_logout_services of this SamlIDPSSODescriptor.  # noqa: E501
        :type single_logout_services: list[SamlEndpoint]
        """

        self._single_logout_services = single_logout_services

    @property
    def single_sign_on_services(self):
        """Gets the single_sign_on_services of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The single_sign_on_services of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: list[SamlEndpoint]
        """
        return self._single_sign_on_services

    @single_sign_on_services.setter
    def single_sign_on_services(self, single_sign_on_services):
        """Sets the single_sign_on_services of this SamlIDPSSODescriptor.


        :param single_sign_on_services: The single_sign_on_services of this SamlIDPSSODescriptor.  # noqa: E501
        :type single_sign_on_services: list[SamlEndpoint]
        """

        self._single_sign_on_services = single_sign_on_services

    @property
    def valid_until(self):
        """Gets the valid_until of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The valid_until of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: datetime
        """
        return self._valid_until

    @valid_until.setter
    def valid_until(self, valid_until):
        """Sets the valid_until of this SamlIDPSSODescriptor.


        :param valid_until: The valid_until of this SamlIDPSSODescriptor.  # noqa: E501
        :type valid_until: datetime
        """

        self._valid_until = valid_until

    @property
    def want_authn_requests_signed(self):
        """Gets the want_authn_requests_signed of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The want_authn_requests_signed of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: bool
        """
        return self._want_authn_requests_signed

    @want_authn_requests_signed.setter
    def want_authn_requests_signed(self, want_authn_requests_signed):
        """Sets the want_authn_requests_signed of this SamlIDPSSODescriptor.


        :param want_authn_requests_signed: The want_authn_requests_signed of this SamlIDPSSODescriptor.  # noqa: E501
        :type want_authn_requests_signed: bool
        """

        self._want_authn_requests_signed = want_authn_requests_signed

    @property
    def xml_name(self):
        """Gets the xml_name of this SamlIDPSSODescriptor.  # noqa: E501


        :return: The xml_name of this SamlIDPSSODescriptor.  # noqa: E501
        :rtype: XmlName
        """
        return self._xml_name

    @xml_name.setter
    def xml_name(self, xml_name):
        """Sets the xml_name of this SamlIDPSSODescriptor.


        :param xml_name: The xml_name of this SamlIDPSSODescriptor.  # noqa: E501
        :type xml_name: XmlName
        """

        self._xml_name = xml_name

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, SamlIDPSSODescriptor):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, SamlIDPSSODescriptor):
            return True

        return self.to_dict() != other.to_dict()



import os
import sys
from typing import List

project_root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
# this line should be on the top to run it on a local environment as 'python vessl/cli/_main.py'
sys.path.append(project_root)

import warnings

warnings.filterwarnings(action='ignore', module='.*paramiko.*')
warnings.filterwarnings(action='ignore', module='.*requests.*')

import click
import sentry_sdk
from click.decorators import pass_context
from sentry_sdk.integrations.logging import ignore_logger

import vessl
from vessl._version import __VERSION__
from vessl.cli._base import VesslGroup, dd_log_cmd
from vessl.cli.dataset import cli as dataset_cli
from vessl.cli.experiment import cli as experiment_cli
from vessl.cli.experiment import create as experiment_create
from vessl.cli.experiment import logs as experiment_logs
from vessl.cli.hub import cli as hub_cli
from vessl.cli.kernel_cluster import cli as kernel_cluster_cli
from vessl.cli.kernel_image import cli as kernel_image_cli
from vessl.cli.kernel_resource_spec import cli as kernel_resource_spec_cli
from vessl.cli.model import model_cli, model_repository_cli
from vessl.cli.organization import cli as organization_cli
from vessl.cli.pipeline import cli as pipeline_cli
from vessl.cli.project import cli as project_cli
from vessl.cli.run import cli as run_cli
from vessl.cli.serve import alias_cli
from vessl.cli.serve import cli as serve_cli
from vessl.cli.ssh_key import cli as ssh_key_cli
from vessl.cli.storage import storage_cli
from vessl.cli.sweep import cli as sweep_cli
from vessl.cli.volume import cli as volume_cli
from vessl.cli.workspace import cli as workspace_cli
from vessl.experiment import list_github_code_refs
from vessl.run import create_run, wrap_str
from vessl.util.common import safe_cast
from vessl.util.config import DEFAULT_CONFIG_PATH, VesslConfigLoader, create_user_data_collection_file, notified_user_data_collection
from vessl.util.constant import (
    ENABLE_SENTRY,
    EXPERIMENT_WORKING_DIR,
    JUPYTER,
    MISTRAL_7B,
    SENTRY_DSN,
    SSD_1B,
    VEGAART,
    VESSL_ENV,
    WHISPER_V3,
)
from vessl.util.echo import print_info, print_warning
from vessl.util.exception import (
    InvalidOrganizationError,
    InvalidProjectError,
    InvalidTokenError,
    VesslApiException,
)
from vessl.util.git_remote import clone_by_force_reset, clone_codes, clone_codes_v2
from vessl.util.prompt import prompt_choices

# Configure Sentry in production
if VESSL_ENV == "prod" and ENABLE_SENTRY:
    sentry_sdk.init(
        SENTRY_DSN,
        traces_sample_rate=1.0,
        ignore_errors=[VesslApiException],
    )
    sentry_sdk.set_tag("cli_version", __VERSION__)
    ignore_logger("vessl.util.logger")
else:
    print_warning(f"Sentry is disabled - VESSL_ENV: {VESSL_ENV}, ENABLE_SENTRY: {ENABLE_SENTRY}")


def prompt_organizations() -> str:
    organizations = vessl.list_organizations()
    organization_count = len(organizations)
    if organization_count == 1:
        return organizations[0].name

    new_organization_string = "Create new organization..."
    choices = [(x.name, i) for i, x in enumerate(organizations)] + [
        (new_organization_string, organization_count)
    ]
    choice = prompt_choices("Default organization", choices)

    if choice == organization_count:
        organization_name = click.prompt("Organization name", type=click.STRING)
        vessl.create_organization(organization_name)
    else:
        organization_name = organizations[choice].name

    return organization_name


_insecure_warning_logged = False

@click.command(cls=VesslGroup)
@click.version_option()
@pass_context
def cli(ctx: click.Context):
    global _insecure_warning_logged
    allow_insecure = safe_cast(os.environ.get('VESSL_INSECURE_SKIP_TLS_VERIFY'), bool, False)
    if allow_insecure and not _insecure_warning_logged:
        vessl.logger.warning('WARNING: VESSL_INSECURE_SKIP_TLS_VERIFY is enabled, HTTPS certificate verification is disabled. Connections may be insecure and vulnerable to man-in-the-middle attacks. Use only in trusted environments such as air-gapped systems or during development.')
        _insecure_warning_logged = True
    vessl.EXEC_MODE = "CLI"
    ctx.ensure_object(dict)


@cli.group(cls=VesslGroup, invoke_without_command=True)
@click.pass_context
@click.option("-t", "--access-token", type=click.STRING)
@click.option("-o", "--organization", type=click.STRING)
@click.option("-p", "--project", type=click.STRING)
@click.option("-f", "--credentials-file", type=click.STRING)
@click.option("--renew-token", is_flag=True)
@click.option("--reset", is_flag=True)
def configure(
    ctx,
    access_token: str,
    organization: str,
    project: str,
    credentials_file: str,
    renew_token: bool,
    reset: bool,
):
    if ctx.invoked_subcommand:
        return

    if vessl.vessl_api.is_in_run_exec_context():
        print("This environment is currently connected to VESSL with VESSL Run context.")
        print("You can manually configure VESSL SDK, but it will break some run functions (e.g. vessl.log())")
        print("Please refer https://docs.vessl.ai/api-reference/cli/getting-started#configuration-precedence.")
        return

    if reset:
        vessl.vessl_api.config_loader = VesslConfigLoader()
        vessl.vessl_api.config_loader.reset()

    try:
        vessl.configure_access_token(
            access_token=access_token,
            credentials_file=credentials_file,
            force_update=renew_token,
        )
    except InvalidTokenError:
        vessl.configure_access_token(force_update=True)

    try:
        vessl.configure_organization(
            organization_name=organization,
            credentials_file=credentials_file,
        )
    except InvalidOrganizationError:
        organization_name = prompt_organizations()
        vessl.configure_organization(organization_name)

    try:
        vessl.configure_project(
            project_name=project,
            credentials_file=credentials_file,
        )
    except InvalidProjectError:
        projects = vessl.list_projects()
        if len(projects) == 1:
            project_name = projects[0].name
        else:
            new_project_string = "Create new project..."
            project_count = len(projects)
            choices = [(x.name, i) for i, x in enumerate(projects)] + [
                (new_project_string, project_count)
            ]
            choice = prompt_choices("Default project", choices)

        if choice == project_count:
            project_name = click.prompt("Project name", type=click.STRING)
            vessl.create_project(project_name)
        else:
            project_name = projects[choice].name

        if project_name is not None:
            vessl.configure_project(project_name)

    print(f"Welcome, {vessl.vessl_api.user.display_name}!")


@cli.group(cls=VesslGroup, invoke_without_command=True)
def whoami():
    if not notified_user_data_collection():
        print_info("Basic information on command executions will be collected to improve future CLI version releases.")
        print_info("More information can be found in ~/.vessl/tracing-agreement file.")
        print()
        create_user_data_collection_file()
    
    if vessl.vessl_api.is_in_run_exec_context():
        vessl.vessl_api.set_access_token(no_prompt=True)
        user = vessl.vessl_api.user
        organization_name = vessl.vessl_api.set_organization()
        project_name = vessl.vessl_api.set_project()
    else:
        config = VesslConfigLoader()
        user = None
        if config.access_token:
            vessl.vessl_api.api_client.set_default_header(
                "Authorization", f"Token {config.access_token}"
            )

            try:
                user = vessl.vessl_api.get_my_user_info_api()
            except VesslApiException:
                pass

        organization_name = config.default_organization
        project_name = config.default_project

        if user is None or organization_name is None:
            print("Please run `vessl configure` first.")
            return

    print(
        f"""Username: {user.username}
Email: {user.email}
Default organization: {organization_name}
Default project: {project_name or 'N/A'}

(The default organization and project can be updated with `vessl configure --reset`.)"""
    )


@configure.vessl_command()
@click.argument("organization", type=click.STRING, required=False)
def organization(organization: str):
    if organization is None:
        organization = prompt_organizations()
    vessl.configure_organization(organization)
    print(f"Saved to {DEFAULT_CONFIG_PATH}.")


@configure.vessl_command()
@click.argument("project", type=click.STRING, required=False)
def project(project: str):
    vessl.vessl_api.set_organization()

    if project is None:
        projects = vessl.list_projects()
        if len(projects) == 0:
            return

        project = prompt_choices("Default project", [x.name for x in projects])
    vessl.configure_project(project)
    print(f"Saved to {DEFAULT_CONFIG_PATH}.")


@configure.command()
def list():
    config = VesslConfigLoader()

    username = ""
    email = ""
    organization = config.default_organization or ""
    project = config.default_project or ""

    if config.access_token:
        vessl.vessl_api.api_client.set_default_header(
            "Authorization", f"Token {config.access_token}"
        )

        try:
            user = vessl.vessl_api.get_my_user_info_api()
            username = user.username
            email = user.email
        except VesslApiException as e:
            pass

    print(
        f"Username: {username}\n"
        f"Email: {email}\n"
        f"Organization: {organization}\n"
        f"Project: {project}"
    )


@cli.vessl_run_command()
@click.pass_context
def hello(ctx):
    ctx.params = {}
    args = []
    ctx.args = configure.parse_args(ctx, args)
    ctx.forward(configure)

    choices = [
        ("[Mistral-7B] Inference app built with Streamlit.", 1),
        ("[SSD-1B] Inference with Segmind's Stable Diffusion model.", 2),
        ("[Whisper-v3] Small inference on data in librispeech_asr.", 3),
        ("[Segmind VegaRT] Inference with Segmind's VegaRT.", 4),
        ("[JupyterLab] Launch a JupyterLab on GPU instance.", 5),
    ]
    choice = prompt_choices("Choose an example run", choices)
    if choice == 1:
        create_run(
            yaml_file=None, yaml_body=MISTRAL_7B, yaml_file_name="mistral-7b.yaml")
    elif choice == 2:
        create_run(
            yaml_file=None, yaml_body=SSD_1B, yaml_file_name="ssd-1b.yaml")
    elif choice == 3:
        create_run(
            yaml_file=None, yaml_body=WHISPER_V3, yaml_file_name="whisper-v3.yaml")
    elif choice == 4:
        create_run(
            yaml_file=None, yaml_body=VEGAART, yaml_file_name="vegart.yaml")
    elif choice == 5:
        create_run(
            yaml_file=None, yaml_body=JUPYTER, yaml_file_name="jupyter.yaml")


@cli.vessl_command(hidden=True)
def download_code():
    """
    This command is not for users and supposed to run on initContainers of experiment, pipeline, and workspace Pods.
    """
    workload_id = safe_cast(os.environ.get("VESSL_WORKLOAD_ID", None), int)
    if workload_id is None:
        print("Failed to download source codes: no workload id.")
        return

    resp = list_github_code_refs(workload_id)
    if resp.scheme == "v2":
        github_code_refs = resp.results_v2
        if not github_code_refs:
            print("No source code to download.")
            return
        clone_codes_v2(github_code_refs)
    else:
        github_code_refs = resp.results
        if not github_code_refs:
            print("No source code to download.")
            return
        clone_codes(github_code_refs)


@cli.vessl_command(hidden=True)
@click.option(
    "-p",
    "--path",
    type=click.STRING,
    help="target directory to import codes.",
)
def import_code(path: str):
    """
    This command is not for users and supposed to run on VESSL Run workloads.
    """
    resp = vessl.vessl_api.run_execution_get_code_ref_api(path=path)
    clone_by_force_reset(resp)


cli.add_command(dataset_cli)
cli.add_command(experiment_cli)
cli.add_command(kernel_cluster_cli)
cli.add_command(kernel_image_cli)
cli.add_command(kernel_resource_spec_cli)
cli.add_command(model_cli)
cli.add_command(model_repository_cli)
cli.add_command(organization_cli)
cli.add_command(project_cli)
cli.add_command(run_cli)
cli.add_command(serve_cli)
cli.add_command(alias_cli)
cli.add_command(ssh_key_cli)
cli.add_command(sweep_cli)
cli.add_command(volume_cli)
cli.add_command(workspace_cli)
cli.add_command(pipeline_cli)
cli.add_command(hub_cli)
cli.add_command(storage_cli)

if __name__ == "__main__":
    cli()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmHubModelTaskEdges(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'category': 'OrmHubModelTaskCategory',
        'created_by_user': 'OrmUser',
        'histories': 'list[OrmHubModelTaskHistory]',
        'hub_tags': 'list[OrmHubTag]',
        'organization_hub_model_tasks': 'list[OrmOrganizationHubModelTask]',
        'owner_organization': 'OrmOrganization',
        'pipeline_spec': 'OrmPipelineSpec'
    }

    attribute_map = {
        'category': 'category',
        'created_by_user': 'created_by_user',
        'histories': 'histories',
        'hub_tags': 'hub_tags',
        'organization_hub_model_tasks': 'organization_hub_model_tasks',
        'owner_organization': 'owner_organization',
        'pipeline_spec': 'pipeline_spec'
    }

    def __init__(self, category=None, created_by_user=None, histories=None, hub_tags=None, organization_hub_model_tasks=None, owner_organization=None, pipeline_spec=None, local_vars_configuration=None):  # noqa: E501
        """OrmHubModelTaskEdges - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._category = None
        self._created_by_user = None
        self._histories = None
        self._hub_tags = None
        self._organization_hub_model_tasks = None
        self._owner_organization = None
        self._pipeline_spec = None
        self.discriminator = None

        if category is not None:
            self.category = category
        if created_by_user is not None:
            self.created_by_user = created_by_user
        if histories is not None:
            self.histories = histories
        if hub_tags is not None:
            self.hub_tags = hub_tags
        if organization_hub_model_tasks is not None:
            self.organization_hub_model_tasks = organization_hub_model_tasks
        if owner_organization is not None:
            self.owner_organization = owner_organization
        if pipeline_spec is not None:
            self.pipeline_spec = pipeline_spec

    @property
    def category(self):
        """Gets the category of this OrmHubModelTaskEdges.  # noqa: E501


        :return: The category of this OrmHubModelTaskEdges.  # noqa: E501
        :rtype: OrmHubModelTaskCategory
        """
        return self._category

    @category.setter
    def category(self, category):
        """Sets the category of this OrmHubModelTaskEdges.


        :param category: The category of this OrmHubModelTaskEdges.  # noqa: E501
        :type category: OrmHubModelTaskCategory
        """

        self._category = category

    @property
    def created_by_user(self):
        """Gets the created_by_user of this OrmHubModelTaskEdges.  # noqa: E501


        :return: The created_by_user of this OrmHubModelTaskEdges.  # noqa: E501
        :rtype: OrmUser
        """
        return self._created_by_user

    @created_by_user.setter
    def created_by_user(self, created_by_user):
        """Sets the created_by_user of this OrmHubModelTaskEdges.


        :param created_by_user: The created_by_user of this OrmHubModelTaskEdges.  # noqa: E501
        :type created_by_user: OrmUser
        """

        self._created_by_user = created_by_user

    @property
    def histories(self):
        """Gets the histories of this OrmHubModelTaskEdges.  # noqa: E501


        :return: The histories of this OrmHubModelTaskEdges.  # noqa: E501
        :rtype: list[OrmHubModelTaskHistory]
        """
        return self._histories

    @histories.setter
    def histories(self, histories):
        """Sets the histories of this OrmHubModelTaskEdges.


        :param histories: The histories of this OrmHubModelTaskEdges.  # noqa: E501
        :type histories: list[OrmHubModelTaskHistory]
        """

        self._histories = histories

    @property
    def hub_tags(self):
        """Gets the hub_tags of this OrmHubModelTaskEdges.  # noqa: E501


        :return: The hub_tags of this OrmHubModelTaskEdges.  # noqa: E501
        :rtype: list[OrmHubTag]
        """
        return self._hub_tags

    @hub_tags.setter
    def hub_tags(self, hub_tags):
        """Sets the hub_tags of this OrmHubModelTaskEdges.


        :param hub_tags: The hub_tags of this OrmHubModelTaskEdges.  # noqa: E501
        :type hub_tags: list[OrmHubTag]
        """

        self._hub_tags = hub_tags

    @property
    def organization_hub_model_tasks(self):
        """Gets the organization_hub_model_tasks of this OrmHubModelTaskEdges.  # noqa: E501


        :return: The organization_hub_model_tasks of this OrmHubModelTaskEdges.  # noqa: E501
        :rtype: list[OrmOrganizationHubModelTask]
        """
        return self._organization_hub_model_tasks

    @organization_hub_model_tasks.setter
    def organization_hub_model_tasks(self, organization_hub_model_tasks):
        """Sets the organization_hub_model_tasks of this OrmHubModelTaskEdges.


        :param organization_hub_model_tasks: The organization_hub_model_tasks of this OrmHubModelTaskEdges.  # noqa: E501
        :type organization_hub_model_tasks: list[OrmOrganizationHubModelTask]
        """

        self._organization_hub_model_tasks = organization_hub_model_tasks

    @property
    def owner_organization(self):
        """Gets the owner_organization of this OrmHubModelTaskEdges.  # noqa: E501


        :return: The owner_organization of this OrmHubModelTaskEdges.  # noqa: E501
        :rtype: OrmOrganization
        """
        return self._owner_organization

    @owner_organization.setter
    def owner_organization(self, owner_organization):
        """Sets the owner_organization of this OrmHubModelTaskEdges.


        :param owner_organization: The owner_organization of this OrmHubModelTaskEdges.  # noqa: E501
        :type owner_organization: OrmOrganization
        """

        self._owner_organization = owner_organization

    @property
    def pipeline_spec(self):
        """Gets the pipeline_spec of this OrmHubModelTaskEdges.  # noqa: E501


        :return: The pipeline_spec of this OrmHubModelTaskEdges.  # noqa: E501
        :rtype: OrmPipelineSpec
        """
        return self._pipeline_spec

    @pipeline_spec.setter
    def pipeline_spec(self, pipeline_spec):
        """Sets the pipeline_spec of this OrmHubModelTaskEdges.


        :param pipeline_spec: The pipeline_spec of this OrmHubModelTaskEdges.  # noqa: E501
        :type pipeline_spec: OrmPipelineSpec
        """

        self._pipeline_spec = pipeline_spec

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmHubModelTaskEdges):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmHubModelTaskEdges):
            return True

        return self.to_dict() != other.to_dict()



import os
import re
import sys
import threading
import time
from typing import Any, Dict, List, Optional, Union

from urllib3.exceptions import MaxRetryError

from vessl.openapi_client.models.experiment_metric_entry import ExperimentMetricEntry
from vessl.openapi_client.models.experiment_metrics_update_api_input import (
    ExperimentMetricsUpdateAPIInput,
)
from vessl.openapi_client.models.response_experiment_info import ResponseExperimentInfo
from vessl.openapi_client.models.run_execution_metrics_update_api_input import (
    RunExecutionMetricsUpdateAPIInput,
)
from vessl.openapi_client.models.v1_plot_entry import V1PlotEntry
from vessl.internal.collector import (
    Collector,
    RunExecutionMetricCollector,
    UserMetricCollector,
)
from vessl.internal.progress_updater import ExecutionProgressUpdater, ProgressUpdater
from vessl.internal.vessl_hyperparameters import (
    HyperparametersController,
)
from vessl.util import constant, logger
from vessl.util.api import VesslApi
from vessl.util.audio import Audio
from vessl.util.constant import (
    VESSL_MEDIA_PATH,
    VESSL_PLOTS_FILETYPE_AUDIO,
    VESSL_PLOTS_FILETYPE_IMAGE,
)
from vessl.util.exception import (
    InvalidTypeError,
    VesslApiException,
    suppress_sdk_exception,
)
from vessl.util.file_watch import WatchJWT
from vessl.util.image import Image

MODE_TEST = "TEST"
MODE_NOT_STARTED = "NOT_STARTED"
MODE_MANAGED = "MANAGED"

SEND_INTERVAL_IN_SEC = 10

METRIC_KEY_REGEX = re.compile("^[a-zA-Z0-9/_-]+$")


class Sender(object):
    def __init__(self, api: VesslApi, experiment_id: int, collectors: List[Collector]):
        self._api = api
        self._experiment_id: int = experiment_id
        self._thread = threading.Thread(target=self._thread_body, daemon=True)
        self._exit = threading.Event()
        self._collectors = collectors

    def stop(self):
        for c in self._collectors:
            c.stop()

        self._exit.set()
        self._thread.join()

    def start(self):
        for c in self._collectors:
            c.start()
        self._thread.start()

    def _thread_body(self):
        while not self._exit.is_set():
            self._send()
            self._exit.wait(timeout=SEND_INTERVAL_IN_SEC)
        self._send()

    def _send(self):
        pairs = [(c, c.collect()) for c in self._collectors]
        for c, m in pairs:
            logger.debug(f"{c} / {m}", str(c), len(m))
        payload = [m for _, metrics in pairs for m in metrics]
        logger.debug(f"Sending {len(payload)} payloads")

        try:
            res = self._api.experiment_metrics_update_api(
                self._experiment_id,
                experiment_metrics_update_api_input=ExperimentMetricsUpdateAPIInput(
                    metrics=payload
                ),
            )
            if res.rejected:
                logger.warning(f"{res.rejected} payloads(s) were rejected.")

            for c, m in pairs:
                c.truncate(len(m))

        except (MaxRetryError, VesslApiException) as e:
            logger.exception("Failed to send metrics to server", exc_info=e)
        except Exception as e:
            logger.exception("Unexpected error", exc_info=e)


class VesslRun(object):
    class ExitHook(object):
        def __init__(self, orig_exit):
            self.orig_exit = orig_exit
            self.exit_code = 0

        def exit(self, code=0):
            self.exit_code = code
            self.orig_exit(code)

    __slots__ = [
        "api",
        "_mode",
        "_collectors",
        "_sender",
        "_progress_updater",
        "hyperparameters",
        "_experiment",
        "_execution",
        "_execution_volume_id",
        "_logger",
        "_user_metric_collector",
        "_run_execution_metric_collector",
        "_exit_hook",
        "_tensorboard_collector",
    ]

    def __init__(self) -> None:
        self.api = VesslApi()

        self._experiment = None
        self._execution = None

        if self.api.is_in_run_exec_context():
            self._execution = True
            self.hyperparameters = None
            self._mode = MODE_MANAGED
            self._set_plot_volume_id()
            self._run_execution_metric_collector = RunExecutionMetricCollector()

            w = WatchJWT(
                path=constant.VESSL_JWT_DIR, logger=logger, callback_function=self._update_jwt
            )
            t = threading.Thread(name="child process", target=w.run, daemon=True)
            t.start()
        else:
            self._experiment = self._get_experiment_from_environment()
            self.hyperparameters = HyperparametersController(self.api, self._experiment)
            self._mode = MODE_NOT_STARTED if self._experiment is None else MODE_MANAGED
            self._user_metric_collector = UserMetricCollector()

        self._exit_hook = self.ExitHook(sys.exit)

    def _set_plot_volume_id(self) -> bool:
        """
        also provides the plot volume id
        """
        try:
            self._execution_volume_id = self.api.run_execution_get_plot_volume_api().volume_id
            return True
        except:
            return False

    def _update_jwt(self, refreshed_token):
        logger.debug("update_jwt to", refreshed_token)
        self.api.reconfigure_access_token_on_jwt_refresh(refreshed_token)

    def _get_experiment_from_environment(self) -> Optional[ResponseExperimentInfo]:
        """Detect experiment from environment variables

        In a Vessl-managed experiment, these variables will be defined.
        """
        experiment_id = os.environ.get("VESSL_EXPERIMENT_ID", None)
        access_token = os.environ.get("VESSL_ACCESS_TOKEN", None)

        if experiment_id is None or access_token is None:
            return None

        self.api.configure_access_token(access_token)
        try:
            return self.api.experiment_read_by_idapi(experiment_id=experiment_id)
        except VesslApiException:
            return None

    def _signal_handler(self, signo, frames):
        sys.exit(130)  # job was terminated by the owner

    def _send_without_collector(self, payloads):
        """Send metrics without using the collector

        In a Vessl-managed experiment, metrics are sent immediately instead of being queued.
        """
        assert self._mode == MODE_MANAGED
        res = self.api.experiment_metrics_update_api(
            self._experiment.id,
            experiment_metrics_update_api_input=ExperimentMetricsUpdateAPIInput(metrics=payloads),
        )
        if res.rejected:
            logger.warning(f"{res.rejected} payloads(s) were rejected.")

    def _send_without_collector_execution(self, payloads):
        """Send metrics without using the collector

        In a Vessl-managed execution, metrics are sent immediately instead of being queued.
        """
        assert self._mode == MODE_MANAGED
        res = self.api.run_execution_metrics_update_api(
            run_execution_metrics_update_api_input=RunExecutionMetricsUpdateAPIInput(
                metrics=payloads
            ),
        )
        if res.rejected:
            logger.warning(f"{res.rejected} payloads(s) were rejected.")

    def init(
        self,
        message: str = None,
        hp: dict = None,
        **kwargs,
    ):
        """Main function to setup Vessl in a local setting

        If this is a Vessl-managed experiment or vessl.init has already been called,
        this will do nothing.

        Args:
            message (str): experiment message
            hp (dict): hyperparameters
        """
        logger.warning("vessl.init() will be deprecated. Use VESSL Run instead.")

        if self._execution:
            print("Connected to VESSL Run.")
            return

        if self._mode == MODE_MANAGED:
            if hp is not None:
                self.hyperparameters.update(hp)

            if message is not None and self._experiment.message != message:
                from vessl.experiment import update_experiment

                self._experiment = update_experiment(
                    self._experiment.number,
                    message=message,
                    organization_name=self._experiment.organization.name,
                    project_name=self._experiment.project.name,
                )
                logger.debug(f'Updated experiment.message to: "{message}"')
            else:
                logger.debug(
                    f'Connected to experiment #{self._experiment.number}. Message: "{self._experiment.message}"'
                )

    def finish(self):
        logger.warning("vessl.finish() will be deprecated. Use VESSL Run instead.")
        return

    @suppress_sdk_exception
    def log(
        self,
        payload: Dict[str, Any],
        step: Optional[int] = None,
        ts: Optional[float] = None,
    ):
        """Log metrics to Vessl

        Args:
            payload (Dict[str, Any]): to log a scalar, value should be a number. To
                log an image, pass a single image or a list of images (type `vessl.util.image.Image`).
            step (int): step.
        """
        if self._execution:
            if self._mode == MODE_NOT_STARTED:
                logger.warning("Invalid usage. This method is only available in VESSL managed environment.")
                return
            if ts is None:
                ts = time.time()

            log_dict = {}
            scalar_dict = {}
            media_dict = {}

            for k, v in payload.items():
                if isinstance(v, list):
                    if all(isinstance(i, Image) for i in v) or all(isinstance(i, Audio) for i in v):
                        media_dict[k] = v
                elif isinstance(v, Image) or isinstance(v, Audio):
                    media_dict[k] = [v]
                elif isinstance(v, int) or isinstance(v, float):
                    scalar_dict[k] = v
                elif isinstance(v, str):
                    log_dict[k] = v
                else:
                    try:
                        scalar_dict[k] = float(v)
                    except ValueError:
                        logger.error(
                            f"Could not convert '{type(v).__name__}' {v} to float. "
                            f"(allowed types: Image, Audio, int, float, str)"
                        )

            if scalar_dict or step is not None:
                self._run_execution_metric_collector.handle_step(step)

            if media_dict:
                self._update_media_execution(media_dict, ts, self._execution_volume_id, step)

            if scalar_dict:
                self._update_metrics_execution(scalar_dict, ts)

            if log_dict:
                self._update_logs_execution(log_dict, ts)

        if self._experiment is not None:
            if self._mode == MODE_NOT_STARTED:
                logger.warning("Invalid usage. This method is only available in VESSL managed environment.")
                return

            if ts is None:
                ts = time.time()

            scalar_dict = {}
            media_dict = {}

            for k, v in payload.items():
                if isinstance(v, list):
                    if all(isinstance(i, Image) for i in v) or all(isinstance(i, Audio) for i in v):
                        media_dict[k] = v
                elif isinstance(v, Image) or isinstance(v, Audio):
                    media_dict[k] = [v]
                elif isinstance(v, int) or isinstance(v, float):
                    scalar_dict[k] = v
                else:
                    try:
                        scalar_dict[k] = float(v)
                    except ValueError:
                        logger.error(
                            f"Could not convert '{type(v).__name__}' {v} to float. "
                            f"(allowed types: Image, Audio, int, float)"
                        )

            # Update step if step is specified. If a scalar is defined but step wasn't,
            # step will be autoincremented.
            if scalar_dict or step is not None:
                self._user_metric_collector.handle_step(step)

            if media_dict:
                self._update_media(media_dict, ts)

            if scalar_dict:
                self._update_metrics(scalar_dict, ts)

    # This should mirror app/influx/metric_schema.go > `isValidMetricKey`
    def _is_metric_key_valid(self, key: str):
        if not METRIC_KEY_REGEX.match(key):
            return False
        if key.startswith("/") or key.endswith("/"):
            return False
        for i in range(1, len(key)):
            if key[i] == "/" and key[i - 1] == "/":
                return False
        return True

    def _update_metrics(self, payload: Dict[str, Any], ts: float):
        invalid_keys = [k for k in payload.keys() if not self._is_metric_key_valid(k)]
        if invalid_keys:
            logger.warning(
                f"Invalid metric keys: {' '.join(invalid_keys)}. This payload will be rejected."
            )

        payloads = [self._user_metric_collector.build_metric_payload(payload, ts)]
        if self._mode == MODE_MANAGED:
            self._send_without_collector(payloads)
            return self._user_metric_collector.step
        return self._user_metric_collector.log_metrics(payloads)

    def _update_media(
        self,
        payload: Dict[str, Union[List[Image], List[Audio]]],
        ts: float,
    ):
        media_type, set_media_type = None, False
        path_to_caption = {}
        for media in payload.values():
            for medium in media:
                path = os.path.basename(medium.path)
                path_to_caption[path] = medium.caption
                if not set_media_type:
                    if not isinstance(medium, Image) and not isinstance(medium, Audio):
                        raise InvalidTypeError(f"Invalid payload type error: {type(medium)}")
                    else:
                        media_type = type(medium).__name__
                        set_media_type = True

        from vessl.volume import copy_volume_file

        files = copy_volume_file(
            source_volume_id=None,
            source_path=os.path.join(VESSL_MEDIA_PATH, "."),
            dest_volume_id=self._experiment.experiment_plot_volume,
            dest_path="/",
        )

        for media in payload.values():
            for medium in media:
                medium.flush()

        plot_files = []
        if files:
            plot_files = [
                {
                    "step": None,
                    "path": file.get("path"),
                    "caption": path_to_caption[file.get("path")],
                    "timestamp": ts,
                }
                for file in files
                if file.get("path") in path_to_caption
            ]

        payloads: List[ExperimentMetricEntry] = []
        for f in plot_files:
            payload = {}
            if media_type == Image.__name__:
                payload = {VESSL_PLOTS_FILETYPE_IMAGE: f}
            elif media_type == Audio.__name__:
                payload = {VESSL_PLOTS_FILETYPE_AUDIO: f}
            payloads.append(self._user_metric_collector.build_media_payload(payload, ts))

        if self._mode == MODE_MANAGED:
            self._send_without_collector(payloads)
        else:
            self._user_metric_collector.log_media(payloads)

    def _update_logs_execution(self, log_dict: Dict[str, Any], ts: float):
        payload = V1PlotEntry(
            measurement="log",
            ts=ts,
            tags=None,
            fields=log_dict,
        )
        payloads = [payload]
        self._send_without_collector_execution(payloads)

    def _update_metrics_execution(self, payload: Dict[str, Any], ts: float):
        invalid_keys = [k for k in payload.keys() if not self._is_metric_key_valid(k)]
        if invalid_keys:
            logger.warning(
                f"Invalid metric keys: {' '.join(invalid_keys)}. This payload will be rejected."
            )

        payloads = [self._run_execution_metric_collector.build_metric_payload(payload, ts)]
        if self._mode == MODE_MANAGED:
            self._send_without_collector_execution(payloads)
            return self._run_execution_metric_collector.step
        return self._run_execution_metric_collector.log_metrics(payloads)

    def _update_media_execution(
        self,
        payload: Dict[str, Union[List[Image], List[Audio]]],
        ts: float,
        volume_id: int,
        step: Optional[int] = None,
    ):
        media_type, set_media_type = None, False
        path_to_caption = {}
        for media in payload.values():
            for medium in media:
                if media_type is not None and media_type != type(medium).__name__:
                    raise InvalidTypeError(f"Invalid multiple payload type error: {type(medium)}")

                path = os.path.basename(medium.path)
                if medium.caption == "" or medium.caption is None:
                    path_to_caption[path] = path
                else:
                    path_to_caption[path] = medium.caption

                if not set_media_type:
                    if not isinstance(medium, Image) and not isinstance(medium, Audio):
                        raise InvalidTypeError(f"Invalid payload type error: {type(medium)}")
                    else:
                        media_type = type(medium).__name__
                        set_media_type = True

        from vessl.volume import copy_volume_file

        files = copy_volume_file(
            source_volume_id=None,
            source_path=os.path.join(VESSL_MEDIA_PATH, "."),
            dest_volume_id=volume_id,
            dest_path="/",
        )

        for media in payload.values():
            for medium in media:
                medium.flush()

        plot_files = []
        if files:
            plot_files = [
                {
                    "step": step,
                    "dest_path": file.get("path"),
                    "caption": path_to_caption[file.get("path")],
                }
                for file in files
                if file.get("path") in path_to_caption
            ]

        payloads: List[V1PlotEntry] = []
        for f in plot_files:
            payload = {}
            if media_type == Image.__name__:
                payload = {VESSL_PLOTS_FILETYPE_IMAGE: f}
            elif media_type == Audio.__name__:
                payload = {VESSL_PLOTS_FILETYPE_AUDIO: f}
            payloads.append(self._run_execution_metric_collector.build_media_payload(payload, ts))

        if self._mode == MODE_MANAGED:
            self._send_without_collector_execution(payloads)
        else:
            self._run_execution_metric_collector.log_media(payloads)

    @suppress_sdk_exception
    def progress(self, value: Union[float, int]):
        """Update experiment or execution progress

        Args:
            value (float): progress value as a decimal between 0 and 1
            value (int): progress value for execution between 0 and 1
        """
        if self._execution is None and self._experiment is None:
            logger.warning("Invalid usage. This method is only available in VESSL managed environment.")
            return

        if type(value) == float and not 0 < value <= 1:
            logger.warning(f"Invalid progress value {value}. (0 < value <= 1)")
            return
        if type(value) == int and not 0 < value <= 100:
            logger.warning(f"Invalid progress value {value}. (0 < value <= 100)")
            return

        # Do not initialize in init() since ProgressUpdater might not be used at all
        if self._execution:
            if not hasattr(self, "_progress_updater"):
                self._progress_updater = ExecutionProgressUpdater(self.api)
                self._progress_updater.start()

            if type(value) == float:
                value = int(value * 100)
            self._progress_updater.update(value)
            logger.debug(f"Run progress: {value}")

        if self._experiment is not None:
            if not hasattr(self, "_progress_updater"):
                self._progress_updater = ProgressUpdater(self.api, self._experiment.id)
                self._progress_updater.start()

            self._progress_updater.update(value)
            logger.debug(f"Experiment progress: {value}")



import os
import time

import click

from vessl import vessl_api
from vessl.cli._base import VesslGroup, vessl_argument
from vessl.cli._util import (
  print_data,
  print_logs,
  print_table,
  truncate_datetime,
)
from vessl.util.prompt import prompt_text
from vessl.util.echo import print_info, print_success, print_error
from vessl.cli.organization import organization_name_option
from vessl.cli.project import project_name_option
from vessl.run import (
    create_run,
    delete_run,
    list_run_logs,
    list_runs,
    read_run,
    terminate_run,
    update_run,
    wrap_str,
)
from vessl.util.ssh import ssh_private_key_path_callback, ssh_command_from_endpoint


def prompt_run_description(ctx: click.Context, param: click.Parameter, description: str):
    return prompt_text("Description to change", False)


@click.command(name="run", cls=VesslGroup)
def cli():
    pass


@cli.vessl_command()
@vessl_argument(
    "run_id",
    type=click.INT,
    required=True,
)
@organization_name_option
@project_name_option
def read(run_id: int):
    run = read_run(run_id=run_id)

    resource = "None"
    if run.run_spec.kernel_resource_spec_from_preset:
        run_resource_spec = run.run_spec.kernel_resource_spec_from_preset
        resource = {
            "Name": run_resource_spec.name,
            "CPU Type": run_resource_spec.cpu_type,
            "CPU Limit": run_resource_spec.cpu_limit,
            "Memory Limit": run_resource_spec.memory_limit,
            "GPU Type": run_resource_spec.gpu_type,
            "GPU Limit": run_resource_spec.gpu_limit,
        }

    image = "None"
    if run.run_spec.kernel_image:
        run_kernel_image = run.run_spec.kernel_image
        image = {
            "Name": run_kernel_image.name,
            "URL": run_kernel_image.image_url,
        }

    commands = []
    if run.run_spec.spec.command and len(run.run_spec.spec.command) != 0:
        for c in run.run_spec.spec.command:
            if c.wait:
                commands.append({"wait": c.wait})

            else:
                if c.workdir:
                    commands.append({"workdir": c.workdir, "command": c.command})
                else:
                    commands.append({"command": c.command})

    interactive = False
    if run.run_spec.spec.interactive:
        interactive = {
            "max_runtime": run.run_spec.spec.interactive.max_runtime,
            "jupyter_idle_timeout": run.run_spec.spec.interactive.jupyter.idle_timeout,
        }

    ports = []
    if run.run_spec.spec.ports and len(run.run_spec.spec.ports) != 0:
        run_ports = run.run_spec.spec.ports
        for p in run_ports:
            ports.append(
                {
                    "name": p.name,
                    "number": p.target_port,
                    "protocol": p.protocol,
                }
            )

    env_vars = []
    if run.run_spec.spec.env_vars and len(run.run_spec.spec.env_vars):
        run_env_vars = run.run_spec.spec.env_vars
        for v in run_env_vars:
            env_vars.append(
                {
                    "key": v.key,
                    "value": v.default,
                    "secret": v.secret,
                }
            )

    service_account = "None"
    if run.run_spec.spec.service_account_name != "":
        service_account = run.run_spec.spec.service_account_name

    print_data(
        {
            "ID": run.id,
            "Name": run.run_spec.title,
            "Description": run.run_spec.description,
            "Status": run.status,
            "Status reason": run.status_reason,
            "Created": truncate_datetime(run.create_dt),
            "Cluster": run.run_spec.cluster.name,
            "Resource": resource,
            "Image": image,
            "Start commands": commands,
            "Interactive": interactive,
            "Environment variables": env_vars,
            "Ports": ports,
            "Service account": service_account,
            "Termination protection": run.run_spec.spec.termination_protection,
        }
    )


@cli.vessl_command()
@organization_name_option
@project_name_option
def list():
    runs = list_runs()
    print_table(
        runs,
        ["ID", "Name", "Type", "Status", "Created", "Description"],
        lambda x: [
            x.id,
            x.run_spec.title,
            x.run_spec.type,
            x.status,
            truncate_datetime(x.create_dt),
            x.message,
        ],
    )


@cli.vessl_command()
@organization_name_option
@project_name_option
@click.option(
    "-f",
    "--file",
    type=click.STRING,
    help="YAML file for Run definition.",
)
@click.option(
    "-w",
    "--watch",
    is_flag=True,
    default=False,
    type=click.BOOL,
    help="Wait for the Run to be scheduled and then stream logs to the console.",
)
def create(file: str, watch: bool):
    if not file:
        print("Missing argument")
        return
    if not os.path.exists(file):
        print(wrap_str(f" YAML does not exist! Check YAML path {file}.", "red"))
        return
    with open(file, "r") as yaml_file:
        create_run(yaml_file=yaml_file, yaml_body="", yaml_file_name=file, watch=watch)


@cli.vessl_command()
@vessl_argument(
    "run_id",
    type=click.INT,
    required=True,
)
@vessl_argument("description", type=click.STRING, required=True, prompter=prompt_run_description)
@organization_name_option
@project_name_option
def update(run_id: int, description: str):
    update_run(run_id, description)
    print_success(f"Updated #{run_id}.\n")


@cli.vessl_command()
@vessl_argument(
    "run_id",
    type=click.INT,
    required=True,
)
@click.option(
    "--tail",
    type=click.INT,
    default=200,
    help="Number of lines to display (from the end).",
)
@click.option(
    "-f",
    "--follow",
    is_flag=True,
    default=False,
)
@organization_name_option
@project_name_option
def logs(run_id: int, tail: int, follow: bool):
    if not follow:
        logs = list_run_logs(run_id=run_id, tail=tail)
        print_logs(logs)
        print_info(f"Displayed last {len(logs)} lines of #{run_id}.")
        return

    after = 0
    run_finished_dt = None
    while True:
        if read_run(run_id).status not in ["pending", "running"] and run_finished_dt is None:
            run_finished_dt = time.time()

        if run_finished_dt is not None and (time.time() - run_finished_dt) > 5:
            break

        logs = list_run_logs(
            run_id=run_id,
            before=int(time.time() - 5),
            after=after,
        )
        print_logs(logs)
        if len(logs) > 0:
            after = logs[-1].timestamp + 0.000001

        time.sleep(3)


@cli.vessl_command()
@vessl_argument(
    "run_id",
    type=click.INT,
    required=True,
)
@organization_name_option
@project_name_option
def terminate(run_id: int):
    run = terminate_run(run_id=run_id)
    print_success(f"Terminated '#{run.id}'.\n")


@cli.vessl_command()
@vessl_argument(
    "run_id",
    type=click.INT,
    required=True,
)
@organization_name_option
@project_name_option
def delete(run_id: int):
    delete_run(run_id)
    print_success(f"Deleted #{run_id}.\n")


@cli.vessl_command()
@vessl_argument(
    "run_id",
    type=click.INT,
    required=True,
)
@click.option(
    "-p",
    "--key-path",
    type=click.STRING,
    help="SSH private key path.",
    callback=ssh_private_key_path_callback,
)
@organization_name_option
@project_name_option
def ssh(run_id: int, key_path: str):
    run = read_run(run_id=run_id)

    if not run.status == "running":
        print_error(f"SSH to run failed: Run {run_id} (status: {run.status}) is not running.")
        return

    cmd = ssh_command_from_endpoint(run.endpoints.ssh.endpoint, key_path)
    print_info(cmd)
    os.system(cmd)



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


try:
    from inspect import getfullargspec
except ImportError:
    from inspect import getargspec as getfullargspec
import pprint
import re  # noqa: F401
import six

from vessl.openapi_client.configuration import Configuration


class OrmPipelineStepExecution(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'assignee_email_addresses': 'dict[str, object]',
        'context': 'dict[str, object]',
        'created_dt': 'datetime',
        'edges': 'OrmPipelineStepExecutionEdges',
        'external_service_endpoint': 'str',
        'id': 'int',
        'immutable_slug': 'str',
        'judged_user_id': 'int',
        'judgment_dt': 'datetime',
        'pipeline_execution_id': 'int',
        'status': 'str',
        'step_id': 'int',
        'step_spec_id': 'int',
        'triggering_pipeline_execution_id': 'int',
        'updated_dt': 'datetime',
        'variables': 'dict[str, object]'
    }

    attribute_map = {
        'assignee_email_addresses': 'assignee_email_addresses',
        'context': 'context',
        'created_dt': 'created_dt',
        'edges': 'edges',
        'external_service_endpoint': 'external_service_endpoint',
        'id': 'id',
        'immutable_slug': 'immutable_slug',
        'judged_user_id': 'judged_user_id',
        'judgment_dt': 'judgment_dt',
        'pipeline_execution_id': 'pipeline_execution_id',
        'status': 'status',
        'step_id': 'step_id',
        'step_spec_id': 'step_spec_id',
        'triggering_pipeline_execution_id': 'triggering_pipeline_execution_id',
        'updated_dt': 'updated_dt',
        'variables': 'variables'
    }

    def __init__(self, assignee_email_addresses=None, context=None, created_dt=None, edges=None, external_service_endpoint=None, id=None, immutable_slug=None, judged_user_id=None, judgment_dt=None, pipeline_execution_id=None, status=None, step_id=None, step_spec_id=None, triggering_pipeline_execution_id=None, updated_dt=None, variables=None, local_vars_configuration=None):  # noqa: E501
        """OrmPipelineStepExecution - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._assignee_email_addresses = None
        self._context = None
        self._created_dt = None
        self._edges = None
        self._external_service_endpoint = None
        self._id = None
        self._immutable_slug = None
        self._judged_user_id = None
        self._judgment_dt = None
        self._pipeline_execution_id = None
        self._status = None
        self._step_id = None
        self._step_spec_id = None
        self._triggering_pipeline_execution_id = None
        self._updated_dt = None
        self._variables = None
        self.discriminator = None

        if assignee_email_addresses is not None:
            self.assignee_email_addresses = assignee_email_addresses
        if context is not None:
            self.context = context
        if created_dt is not None:
            self.created_dt = created_dt
        if edges is not None:
            self.edges = edges
        self.external_service_endpoint = external_service_endpoint
        if id is not None:
            self.id = id
        if immutable_slug is not None:
            self.immutable_slug = immutable_slug
        self.judged_user_id = judged_user_id
        self.judgment_dt = judgment_dt
        if pipeline_execution_id is not None:
            self.pipeline_execution_id = pipeline_execution_id
        if status is not None:
            self.status = status
        self.step_id = step_id
        self.step_spec_id = step_spec_id
        self.triggering_pipeline_execution_id = triggering_pipeline_execution_id
        if updated_dt is not None:
            self.updated_dt = updated_dt
        if variables is not None:
            self.variables = variables

    @property
    def assignee_email_addresses(self):
        """Gets the assignee_email_addresses of this OrmPipelineStepExecution.  # noqa: E501


        :return: The assignee_email_addresses of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._assignee_email_addresses

    @assignee_email_addresses.setter
    def assignee_email_addresses(self, assignee_email_addresses):
        """Sets the assignee_email_addresses of this OrmPipelineStepExecution.


        :param assignee_email_addresses: The assignee_email_addresses of this OrmPipelineStepExecution.  # noqa: E501
        :type assignee_email_addresses: dict[str, object]
        """

        self._assignee_email_addresses = assignee_email_addresses

    @property
    def context(self):
        """Gets the context of this OrmPipelineStepExecution.  # noqa: E501


        :return: The context of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._context

    @context.setter
    def context(self, context):
        """Sets the context of this OrmPipelineStepExecution.


        :param context: The context of this OrmPipelineStepExecution.  # noqa: E501
        :type context: dict[str, object]
        """

        self._context = context

    @property
    def created_dt(self):
        """Gets the created_dt of this OrmPipelineStepExecution.  # noqa: E501


        :return: The created_dt of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this OrmPipelineStepExecution.


        :param created_dt: The created_dt of this OrmPipelineStepExecution.  # noqa: E501
        :type created_dt: datetime
        """

        self._created_dt = created_dt

    @property
    def edges(self):
        """Gets the edges of this OrmPipelineStepExecution.  # noqa: E501


        :return: The edges of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: OrmPipelineStepExecutionEdges
        """
        return self._edges

    @edges.setter
    def edges(self, edges):
        """Sets the edges of this OrmPipelineStepExecution.


        :param edges: The edges of this OrmPipelineStepExecution.  # noqa: E501
        :type edges: OrmPipelineStepExecutionEdges
        """

        self._edges = edges

    @property
    def external_service_endpoint(self):
        """Gets the external_service_endpoint of this OrmPipelineStepExecution.  # noqa: E501


        :return: The external_service_endpoint of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: str
        """
        return self._external_service_endpoint

    @external_service_endpoint.setter
    def external_service_endpoint(self, external_service_endpoint):
        """Sets the external_service_endpoint of this OrmPipelineStepExecution.


        :param external_service_endpoint: The external_service_endpoint of this OrmPipelineStepExecution.  # noqa: E501
        :type external_service_endpoint: str
        """

        self._external_service_endpoint = external_service_endpoint

    @property
    def id(self):
        """Gets the id of this OrmPipelineStepExecution.  # noqa: E501


        :return: The id of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this OrmPipelineStepExecution.


        :param id: The id of this OrmPipelineStepExecution.  # noqa: E501
        :type id: int
        """

        self._id = id

    @property
    def immutable_slug(self):
        """Gets the immutable_slug of this OrmPipelineStepExecution.  # noqa: E501


        :return: The immutable_slug of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: str
        """
        return self._immutable_slug

    @immutable_slug.setter
    def immutable_slug(self, immutable_slug):
        """Sets the immutable_slug of this OrmPipelineStepExecution.


        :param immutable_slug: The immutable_slug of this OrmPipelineStepExecution.  # noqa: E501
        :type immutable_slug: str
        """

        self._immutable_slug = immutable_slug

    @property
    def judged_user_id(self):
        """Gets the judged_user_id of this OrmPipelineStepExecution.  # noqa: E501


        :return: The judged_user_id of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: int
        """
        return self._judged_user_id

    @judged_user_id.setter
    def judged_user_id(self, judged_user_id):
        """Sets the judged_user_id of this OrmPipelineStepExecution.


        :param judged_user_id: The judged_user_id of this OrmPipelineStepExecution.  # noqa: E501
        :type judged_user_id: int
        """

        self._judged_user_id = judged_user_id

    @property
    def judgment_dt(self):
        """Gets the judgment_dt of this OrmPipelineStepExecution.  # noqa: E501


        :return: The judgment_dt of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: datetime
        """
        return self._judgment_dt

    @judgment_dt.setter
    def judgment_dt(self, judgment_dt):
        """Sets the judgment_dt of this OrmPipelineStepExecution.


        :param judgment_dt: The judgment_dt of this OrmPipelineStepExecution.  # noqa: E501
        :type judgment_dt: datetime
        """

        self._judgment_dt = judgment_dt

    @property
    def pipeline_execution_id(self):
        """Gets the pipeline_execution_id of this OrmPipelineStepExecution.  # noqa: E501


        :return: The pipeline_execution_id of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: int
        """
        return self._pipeline_execution_id

    @pipeline_execution_id.setter
    def pipeline_execution_id(self, pipeline_execution_id):
        """Sets the pipeline_execution_id of this OrmPipelineStepExecution.


        :param pipeline_execution_id: The pipeline_execution_id of this OrmPipelineStepExecution.  # noqa: E501
        :type pipeline_execution_id: int
        """

        self._pipeline_execution_id = pipeline_execution_id

    @property
    def status(self):
        """Gets the status of this OrmPipelineStepExecution.  # noqa: E501


        :return: The status of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this OrmPipelineStepExecution.


        :param status: The status of this OrmPipelineStepExecution.  # noqa: E501
        :type status: str
        """

        self._status = status

    @property
    def step_id(self):
        """Gets the step_id of this OrmPipelineStepExecution.  # noqa: E501


        :return: The step_id of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: int
        """
        return self._step_id

    @step_id.setter
    def step_id(self, step_id):
        """Sets the step_id of this OrmPipelineStepExecution.


        :param step_id: The step_id of this OrmPipelineStepExecution.  # noqa: E501
        :type step_id: int
        """

        self._step_id = step_id

    @property
    def step_spec_id(self):
        """Gets the step_spec_id of this OrmPipelineStepExecution.  # noqa: E501


        :return: The step_spec_id of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: int
        """
        return self._step_spec_id

    @step_spec_id.setter
    def step_spec_id(self, step_spec_id):
        """Sets the step_spec_id of this OrmPipelineStepExecution.


        :param step_spec_id: The step_spec_id of this OrmPipelineStepExecution.  # noqa: E501
        :type step_spec_id: int
        """

        self._step_spec_id = step_spec_id

    @property
    def triggering_pipeline_execution_id(self):
        """Gets the triggering_pipeline_execution_id of this OrmPipelineStepExecution.  # noqa: E501


        :return: The triggering_pipeline_execution_id of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: int
        """
        return self._triggering_pipeline_execution_id

    @triggering_pipeline_execution_id.setter
    def triggering_pipeline_execution_id(self, triggering_pipeline_execution_id):
        """Sets the triggering_pipeline_execution_id of this OrmPipelineStepExecution.


        :param triggering_pipeline_execution_id: The triggering_pipeline_execution_id of this OrmPipelineStepExecution.  # noqa: E501
        :type triggering_pipeline_execution_id: int
        """

        self._triggering_pipeline_execution_id = triggering_pipeline_execution_id

    @property
    def updated_dt(self):
        """Gets the updated_dt of this OrmPipelineStepExecution.  # noqa: E501


        :return: The updated_dt of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this OrmPipelineStepExecution.


        :param updated_dt: The updated_dt of this OrmPipelineStepExecution.  # noqa: E501
        :type updated_dt: datetime
        """

        self._updated_dt = updated_dt

    @property
    def variables(self):
        """Gets the variables of this OrmPipelineStepExecution.  # noqa: E501


        :return: The variables of this OrmPipelineStepExecution.  # noqa: E501
        :rtype: dict[str, object]
        """
        return self._variables

    @variables.setter
    def variables(self, variables):
        """Sets the variables of this OrmPipelineStepExecution.


        :param variables: The variables of this OrmPipelineStepExecution.  # noqa: E501
        :type variables: dict[str, object]
        """

        self._variables = variables

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = getfullargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, OrmPipelineStepExecution):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, OrmPipelineStepExecution):
            return True

        return self.to_dict() != other.to_dict()
